import dataclasses
from random import Random
from typing import Tuple, List

import pytest

from randovania.game_description import default_database
from randovania.game_description.game_description import GameDescription
from randovania.game_description.requirements import Requirement
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_info import add_resources_into_another
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.search import find_resource_info_with_long_name
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.game_description.world.area import Area
from randovania.game_description.world.dock import DockWeaknessDatabase
from randovania.game_description.world.node import ResourceNode, GenericNode, TranslatorGateNode
from randovania.game_description.world.world import World
from randovania.game_description.world.world_list import WorldList
from randovania.games.game import RandovaniaGame
from randovania.generator import base_patches_factory
from randovania.generator.generator_reach import (
    GeneratorReach)
from randovania.generator.old_generator_reach import OldGeneratorReach
from randovania.generator.reach_lib import filter_pickup_nodes, collectable_resource_nodes, \
    get_collectable_resource_nodes_of_reach, reach_with_all_safe_resources, advance_reach_with_possible_unsafe_resources
from randovania.generator.item_pool import pool_creator
from randovania.layout.base.base_configuration import StartingLocationList
from randovania.layout.base.trick_level import LayoutTrickLevel
from randovania.layout.base.trick_level_configuration import TrickLevelConfiguration
from randovania.layout.permalink import Permalink
from randovania.layout.preset import Preset
from randovania.resolver import bootstrap
from randovania.resolver.state import State, add_pickup_to_state, StateGameData


def run_bootstrap(preset: Preset):
    game = default_database.game_description_for(preset.game).make_mutable_copy()
    game.resource_database = bootstrap.patch_resource_database(game.resource_database, preset.configuration)
    permalink = Permalink(
        seed_number=15000,
        spoiler=True,
        presets={0: preset},
    )
    patches = base_patches_factory.create_base_patches(preset.configuration, Random(15000), game, False, player_index=0)
    _, state = bootstrap.logic_bootstrap(preset.configuration, game, patches)

    return game, state, permalink


def _create_reach_with_unsafe(game: GameDescription, state: State) -> GeneratorReach:
    return advance_reach_with_possible_unsafe_resources(reach_with_all_safe_resources(game, state))


def _create_reaches_and_compare(game: GameDescription, state: State) -> Tuple[GeneratorReach, GeneratorReach]:
    first_reach = _create_reach_with_unsafe(game, state)
    second_reach = _create_reach_with_unsafe(game, first_reach.state)

    assert first_reach.is_safe_node(first_reach.state.node)
    assert second_reach.is_safe_node(first_reach.state.node)
    assert first_reach.is_safe_node(second_reach.state.node)
    assert second_reach.is_safe_node(second_reach.state.node)

    assert set(first_reach.safe_nodes) == set(second_reach.safe_nodes)
    assert set(first_reach.nodes) == set(second_reach.nodes)

    return first_reach, second_reach


def _compare_actions(first_reach: GeneratorReach,
                     second_reach: GeneratorReach,
                     ) -> Tuple[List[ResourceNode], List[ResourceNode]]:
    first_actions = get_collectable_resource_nodes_of_reach(first_reach)
    second_actions = get_collectable_resource_nodes_of_reach(second_reach)
    assert set(first_actions) == set(second_actions)

    return first_actions, second_actions


@pytest.mark.parametrize(("game_enum", "preset_name", "ignore_events", "ignore_pickups"), [
    (RandovaniaGame.METROID_PRIME_ECHOES, "Starter Preset", {91, 92, 97}, set()),  # Echoes
    (RandovaniaGame.METROID_PRIME_CORRUPTION, "Starter Preset", {1, 146, 147, 148, 154}, {0, 1, 2}),  # Corruption
    (RandovaniaGame.METROID_PRIME, "Starter Preset", {33}, set())  # Prime
])
def test_database_collectable(preset_manager, game_enum, preset_name, ignore_events, ignore_pickups):
    game, initial_state, permalink = run_bootstrap(
        preset_manager.included_preset_with(game_enum, preset_name).get_preset())
    all_pickups = set(filter_pickup_nodes(game.world_list.all_nodes))
    pool_results = pool_creator.calculate_pool_results(permalink.get_preset(0).configuration,
                                                       game.resource_database)
    add_resources_into_another(initial_state.resources, pool_results.initial_resources)
    for pickup in pool_results.pickups:
        add_pickup_to_state(initial_state, pickup)
    for pickup in pool_results.assignment.values():
        add_pickup_to_state(initial_state, pickup)
    for trick in game.resource_database.trick:
        initial_state.resources[trick] = LayoutTrickLevel.maximum().as_number

    expected_events = [event for event in game.resource_database.event if event.index not in ignore_events]
    expected_pickups = sorted(it.pickup_index for it in all_pickups if it.pickup_index.index not in ignore_pickups)

    reach = _create_reach_with_unsafe(game, initial_state.heal())
    while list(collectable_resource_nodes(reach.nodes, reach)):
        reach.act_on(next(iter(collectable_resource_nodes(reach.nodes, reach))))
        reach = advance_reach_with_possible_unsafe_resources(reach)

    # print("\nCurrent reach:")
    # for world in game.world_list.worlds:
    #     print(f"\n>> {world.name}")
    #     for node in world.all_nodes:
    #         print("[{!s:>5}, {!s:>5}, {!s:>5}] {}".format(
    #             reach.is_reachable_node(node), reach.is_safe_node(node),
    #             reach.state.resources.get(node.resource(), 0) > 0 if isinstance(node, ResourceNode) else "",
    #             game.world_list.node_name(node)))

    collected_indices = {
        resource
        for resource, quantity in reach.state.resources.items()
        if quantity > 0 and isinstance(resource, PickupIndex)
    }
    collected_events = {
        resource
        for resource, quantity in reach.state.resources.items()
        if quantity > 0 and resource.resource_type == ResourceType.EVENT
    }
    assert list(collectable_resource_nodes(reach.nodes, reach)) == []
    assert sorted(collected_indices) == expected_pickups
    assert sorted(collected_events, key=lambda it: it.index) == expected_events


@pytest.mark.parametrize("has_translator", [False, True])
def test_basic_search_with_translator_gate(has_translator: bool, echoes_resource_database):
    # Setup
    scan_visor = echoes_resource_database.get_item(10)

    node_a = GenericNode("Node A", True, None, 0)
    node_b = GenericNode("Node B", True, None, 1)
    node_c = GenericNode("Node C", True, None, 2)
    translator_node = TranslatorGateNode("Translator Gate", True, None, 3, TranslatorGate(1), scan_visor)

    world_list = WorldList([
        World("Test World", "Test Dark World", 1, [
            Area("Test Area A", False, 10, 0, True, [node_a, node_b, node_c, translator_node],
                 {
                     node_a: {
                         node_b: Requirement.trivial(),
                         translator_node: Requirement.trivial(),
                     },
                     node_b: {
                         node_a: Requirement.trivial(),
                     },
                     node_c: {
                         translator_node: Requirement.trivial(),
                     },
                     translator_node: {
                         node_a: Requirement.trivial(),
                         node_c: Requirement.trivial(),
                     },
                 }
                 )
        ])
    ])
    game = GameDescription(RandovaniaGame.METROID_PRIME_ECHOES, DockWeaknessDatabase([], [], [], []),
                           echoes_resource_database, Requirement.impossible(),
                           None, {}, None, world_list)

    patches = game.create_game_patches()
    patches = patches.assign_gate_assignment({
        TranslatorGate(1): scan_visor
    })
    initial_state = State({scan_visor: 1 if has_translator else 0}, (), 99,
                          node_a, patches, None, StateGameData(echoes_resource_database, game.world_list, 100))

    # Run
    reach = reach_with_all_safe_resources(game, initial_state)

    # Assert
    if has_translator:
        assert set(reach.safe_nodes) == {node_a, node_b, translator_node, node_c}
    else:
        assert set(reach.safe_nodes) == {node_a, node_b}


def test_reach_size_from_start_echoes(small_echoes_game_description, default_layout_configuration):
    # Setup
    game = small_echoes_game_description
    specific_levels = {
        trick.short_name: LayoutTrickLevel.maximum()
        for trick in game.resource_database.trick
    }

    def item(name: str):
        return find_resource_info_with_long_name(game.resource_database.item, name)

    def nodes(*names):
        result = [
            game.world_list.node_from_name(name)
            for name in names
        ]
        result.sort(key=lambda it: it.index)
        return result

    layout_configuration = dataclasses.replace(
        default_layout_configuration,
        trick_level=TrickLevelConfiguration(minimal_logic=False,
                                            specific_levels=specific_levels,
                                            game=default_layout_configuration.game),
        starting_location=StartingLocationList.with_elements(
            [game.starting_location],
            game=RandovaniaGame.METROID_PRIME_ECHOES,
        )
    )
    patches = base_patches_factory.create_base_patches(layout_configuration, Random(15000), game,
                                                       False, player_index=0)
    state = bootstrap.calculate_starting_state(game, patches, 100)
    state.resources[item("Combat Visor")] = 1
    state.resources[item("Amber Translator")] = 1
    state.resources[item("Scan Visor")] = 1
    state.resources[item("Morph Ball")] = 1
    state.resources[item("Power Beam")] = 1
    state.resources[item("Charge Beam")] = 1
    state.resources[item("Grapple Beam")] = 1
    state.resources[item("Dark Beam")] = 1
    state.resources[item("Dark Ammo")] = 50
    state.resources[item("Missile")] = 5

    # Run
    reach = OldGeneratorReach.reach_from_state(game, state)

    # Assert
    assert list(reach.nodes) == nodes(
        "Temple Grounds/Path of Eyes/Front of Translator Gate",
        "Temple Grounds/Path of Eyes/Lore Scan",
        "Temple Grounds/Path of Eyes/Translator Gate",
        "Temple Grounds/Path of Eyes/Door to Torvus Transport Access",

        "Temple Grounds/Torvus Transport Access/Door to Path of Eyes",
        "Temple Grounds/Torvus Transport Access/Door to Transport to Torvus Bog",

        "Temple Grounds/Transport to Torvus Bog/Door to Torvus Transport Access",
        "Temple Grounds/Transport to Torvus Bog/Elevator to Torvus Bog - Transport to Temple Grounds",

        "Torvus Bog/Transport to Temple Grounds/Elevator to Temple Grounds - Transport to Torvus Bog",
        "Torvus Bog/Transport to Temple Grounds/Door to Temple Transport Access",

        "Torvus Bog/Temple Transport Access/Door to Transport to Temple Grounds",
        "Torvus Bog/Temple Transport Access/Door to Torvus Lagoon",

        "Torvus Bog/Torvus Lagoon/Door to Temple Transport Access",
        "Torvus Bog/Torvus Lagoon/Door to Path of Roots",
        "Torvus Bog/Torvus Lagoon/Keybearer Corpse (S-Dly)",

        "Torvus Bog/Path of Roots/Door to Torvus Lagoon",
        "Torvus Bog/Path of Roots/Door to Great Bridge",
        "Torvus Bog/Path of Roots/Pickup (Missile)",
        "Torvus Bog/Path of Roots/Next to Pickup",
        "Torvus Bog/Path of Roots/Under Lore Scan",
        "Torvus Bog/Path of Roots/Lore Scan",

        "Torvus Bog/Great Bridge/Door to Path of Roots",
    )
    assert len(list(reach.safe_nodes)) == 20



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_elevators.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from randovania.gui.lib.scroll_protected import ScrollProtectedComboBox


class Ui_PresetElevators(object):
    def setupUi(self, PresetElevators):
        if not PresetElevators.objectName():
            PresetElevators.setObjectName(u"PresetElevators")
        PresetElevators.resize(505, 463)
        self.centralWidget = QWidget(PresetElevators)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.elevator_parent_layout = QVBoxLayout(self.centralWidget)
        self.elevator_parent_layout.setSpacing(6)
        self.elevator_parent_layout.setContentsMargins(11, 11, 11, 11)
        self.elevator_parent_layout.setObjectName(u"elevator_parent_layout")
        self.elevator_parent_layout.setContentsMargins(0, 0, 0, 0)
        self.elevator_scroll_area = QScrollArea(self.centralWidget)
        self.elevator_scroll_area.setObjectName(u"elevator_scroll_area")
        self.elevator_scroll_area.setWidgetResizable(True)
        self.elevator_scroll_area_contents = QWidget()
        self.elevator_scroll_area_contents.setObjectName(u"elevator_scroll_area_contents")
        self.elevator_scroll_area_contents.setGeometry(QRect(0, 0, 489, 881))
        self.elevator_layout = QVBoxLayout(self.elevator_scroll_area_contents)
        self.elevator_layout.setSpacing(6)
        self.elevator_layout.setContentsMargins(11, 11, 11, 11)
        self.elevator_layout.setObjectName(u"elevator_layout")
        self.elevator_layout.setContentsMargins(4, 6, 4, 0)
        self.elevators_description_label = QLabel(self.elevator_scroll_area_contents)
        self.elevators_description_label.setObjectName(u"elevators_description_label")
        self.elevators_description_label.setWordWrap(True)

        self.elevator_layout.addWidget(self.elevators_description_label)

        self.elevators_combo = ScrollProtectedComboBox(self.elevator_scroll_area_contents)
        self.elevators_combo.setObjectName(u"elevators_combo")

        self.elevator_layout.addWidget(self.elevators_combo)

        self.elevators_help_sound_bug_label = QLabel(self.elevator_scroll_area_contents)
        self.elevators_help_sound_bug_label.setObjectName(u"elevators_help_sound_bug_label")
        self.elevators_help_sound_bug_label.setWordWrap(True)

        self.elevator_layout.addWidget(self.elevators_help_sound_bug_label)

        self.elevators_line_2 = QFrame(self.elevator_scroll_area_contents)
        self.elevators_line_2.setObjectName(u"elevators_line_2")
        self.elevators_line_2.setFrameShape(QFrame.HLine)
        self.elevators_line_2.setFrameShadow(QFrame.Sunken)

        self.elevator_layout.addWidget(self.elevators_line_2)

        self.skip_final_bosses_check = QCheckBox(self.elevator_scroll_area_contents)
        self.skip_final_bosses_check.setObjectName(u"skip_final_bosses_check")

        self.elevator_layout.addWidget(self.skip_final_bosses_check)

        self.skip_final_bosses_label = QLabel(self.elevator_scroll_area_contents)
        self.skip_final_bosses_label.setObjectName(u"skip_final_bosses_label")
        self.skip_final_bosses_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.skip_final_bosses_label.setWordWrap(True)

        self.elevator_layout.addWidget(self.skip_final_bosses_label)

        self.elevators_line_4 = QFrame(self.elevator_scroll_area_contents)
        self.elevators_line_4.setObjectName(u"elevators_line_4")
        self.elevators_line_4.setFrameShape(QFrame.HLine)
        self.elevators_line_4.setFrameShadow(QFrame.Sunken)

        self.elevator_layout.addWidget(self.elevators_line_4)

        self.elevators_allow_unvisited_names_check = QCheckBox(self.elevator_scroll_area_contents)
        self.elevators_allow_unvisited_names_check.setObjectName(u"elevators_allow_unvisited_names_check")

        self.elevator_layout.addWidget(self.elevators_allow_unvisited_names_check)

        self.elevators_line_3 = QFrame(self.elevator_scroll_area_contents)
        self.elevators_line_3.setObjectName(u"elevators_line_3")
        self.elevators_line_3.setFrameShape(QFrame.HLine)
        self.elevators_line_3.setFrameShadow(QFrame.Sunken)

        self.elevator_layout.addWidget(self.elevators_line_3)

        self.elevators_help_list_label = QLabel(self.elevator_scroll_area_contents)
        self.elevators_help_list_label.setObjectName(u"elevators_help_list_label")
        self.elevators_help_list_label.setWordWrap(True)

        self.elevator_layout.addWidget(self.elevators_help_list_label)

        self.elevators_source_group = QGroupBox(self.elevator_scroll_area_contents)
        self.elevators_source_group.setObjectName(u"elevators_source_group")
        self.elevators_source_layout = QGridLayout(self.elevators_source_group)
        self.elevators_source_layout.setSpacing(3)
        self.elevators_source_layout.setContentsMargins(11, 11, 11, 11)
        self.elevators_source_layout.setObjectName(u"elevators_source_layout")
        self.elevators_source_layout.setContentsMargins(1, 1, 1, 1)

        self.elevator_layout.addWidget(self.elevators_source_group)

        self.elevators_target_group = QGroupBox(self.elevator_scroll_area_contents)
        self.elevators_target_group.setObjectName(u"elevators_target_group")
        self.elevators_target_layout = QGridLayout(self.elevators_target_group)
        self.elevators_target_layout.setSpacing(3)
        self.elevators_target_layout.setContentsMargins(11, 11, 11, 11)
        self.elevators_target_layout.setObjectName(u"elevators_target_layout")
        self.elevators_target_layout.setContentsMargins(1, 1, 1, 1)

        self.elevator_layout.addWidget(self.elevators_target_group)

        self.elevator_scroll_area.setWidget(self.elevator_scroll_area_contents)

        self.elevator_parent_layout.addWidget(self.elevator_scroll_area)

        PresetElevators.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetElevators)

        QMetaObject.connectSlotsByName(PresetElevators)
    # setupUi

    def retranslateUi(self, PresetElevators):
        PresetElevators.setWindowTitle(QCoreApplication.translate("PresetElevators", u"Elevators", None))
        self.elevators_description_label.setText(QCoreApplication.translate("PresetElevators", u"<html><head/><body><p>Controls where each elevator connects to.</p><p><span style=\" font-weight:600;\">Original Connections</span>: all elevators are connected to where they do in the original game.</p><p><span style=\" font-weight:600;\">Two-way, between areas</span>: after taking an elevator, the elevator in the room you are in will bring you back to where you were. An elevator will never connect to another in the same area. This is the only setting that guarantees all areas are reachable.</p><p><span style=\" font-weight:600;\">Two-way, unchecked</span>: after taking an elevator, the elevator in the room you are in will bring you back to where you were.</p><p><span style=\" font-weight:600;\">One-way, elevator room with cycles</span>: all elevators bring you to an elevator room, but going backwards can go somewhere else. All rooms are used as a destination exactly once, causing all elevators to be separated into loops.</p><p><span style=\" font-weight:600;\">One-way, elevator room with replacement</span>: "
                        "all elevators bring you to an elevator room, but going backwards can go somewhere else. Rooms can be used as a destination multiple times, causing elevators which you can possibly not come back to.</p><p><span style=\" font-weight:600;\">One-way, anywhere</span>: elevators are connected to any room from the game.</p></body></html>", None))
        self.elevators_help_sound_bug_label.setText(QCoreApplication.translate("PresetElevators", u"<html><head/><body><p>The looping sound effect that plays during the elevator cutscene only stops playing during the cutscene of the elevator landing, which only exists on the rooms that have an elevator. This also includes Sky Temple Energy Controller when coming from somewhere other than Sky Temple Gateway and Aerie Transport Station before Dark Samus 2 is defeated.</p><p>This bug can be minimized by disabling &quot;Keep Elevator/Teleporter transition sound effect&quot; in the &quot;Customize in-game settings&quot; dialog.</p></body></html>", None))
        self.skip_final_bosses_check.setText(QCoreApplication.translate("PresetElevators", u"Go directly to credits from Sky Temple Gateway", None))
        self.skip_final_bosses_label.setText(QCoreApplication.translate("PresetElevators", u"<html><head/><body><p>Change the light beam in Sky Temple Gateway to go directly to the credits, skipping the final bosses.</p><p>This changes the requirements to <span style=\" font-weight:600;\">not need the final bosses</span>, turning certain items optional such as Screw Attack and Spider Ball.</p></body></html>", None))
        self.elevators_allow_unvisited_names_check.setText(QCoreApplication.translate("PresetElevators", u"Allow \"Always display room names on map\" when elevators are shuffled", None))
        self.elevators_help_list_label.setText(QCoreApplication.translate("PresetElevators", u"<html><head/><body><p>Shuffling Sky Temple Gateway, Sky Temple Energy Controller, Aerie and Aerie Transport Station is possible, but they're not included by default as they behave somewhat differently to other elevators.</p><p>The elevator in Aerie Transport Station is only available after you defeat Dark Samus 2.</p><p>When shuffling Sky Temple Energy Controller, you <span style=\" font-weight:600;\">must</span> enter Sky Temple Gateway via an elevator otherwise the game will crash.</p><p><span style=\" font-style:italic;\">Warning</span>: Entering Sky Temple Energy Controller from elsewhere causes the game to be stuck in a black screen in unknown conditions. The game is still running, so you can blindly use the menu mod to work around this issue.</p></body></html>", None))
        self.elevators_source_group.setTitle(QCoreApplication.translate("PresetElevators", u"Elevators to randomize", None))
        self.elevators_target_group.setTitle(QCoreApplication.translate("PresetElevators", u"Valid elevator targets", None))
    # retranslateUi




import dataclasses

from randovania.gui.lib import preset_describer
from randovania.layout.prime2.echoes_configuration import LayoutSkyTempleKeyMode


def test_echoes_format_params(default_layout_configuration):
    # Setup
    layout_configuration = dataclasses.replace(
        default_layout_configuration,
        sky_temple_keys=LayoutSkyTempleKeyMode.ALL_BOSSES,
    )

    # Run
    result = preset_describer._echoes_format_params(layout_configuration)

    # Assert
    assert result == {
        'Difficulty': [
            'Damage Strictness: Medium',
        ],
        'Game Changes': [
            'Missiles needs Launcher, Power Bomb needs Main',
            'Warp to start, Menu Mod',
        ],
        'Gameplay': [
            'Starting Location: Temple Grounds - Landing Site',
            'Translator Gates: Vanilla (Colors)',
        ],
        'Item Placement': [
            'All tricks disabled',
            'Dangerous Actions: Randomly',
        ],
        'Item Pool': [
            'Size: 118 of 119',
            'Progressive Suit, Split beam ammo',
            'Sky Temple Keys at all bosses',
        ],
        'Starting Items': [
            'Vanilla',
        ]
    }



from randovania.game_description.requirements import ResourceRequirement, RequirementAnd
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.game_description.world.area import Area
from randovania.game_description.world.dock import DockConnection, DockWeakness, DockLockType, DockType
from randovania.game_description.world.node import DockNode
from randovania.game_description.world.world import World
from randovania.game_description.world.world_list import WorldList


def test_connections_from_dock_blast_shield(empty_patches):
    # Setup
    req_1 = ResourceRequirement(SimpleResourceInfo(1, "Ev1", "Ev1", ResourceType.EVENT), 1, False)
    req_2 = ResourceRequirement(SimpleResourceInfo(2, "Ev2", "Ev2", ResourceType.EVENT), 1, False)
    weak_1 = DockWeakness(1, "Weak 1", DockLockType.FRONT_ALWAYS_BACK_FREE, req_1, DockType.DOOR)
    weak_2 = DockWeakness(2, "Weak 2", DockLockType.FRONT_BLAST_BACK_BLAST, req_2, DockType.DOOR)

    node_1 = DockNode("Node 1", False, None, 0, 0, DockConnection(0x30, 0), weak_1)
    node_2 = DockNode("Node 2", False, None, 1, 0, DockConnection(0x20, 0), weak_2)

    area_1 = Area("Area 1", False, 0x20, 0, True, [node_1], {})
    area_2 = Area("Area 2", False, 0x30, 0, True, [node_2], {})

    world = World("World", None, 0x10, [area_1, area_2])
    world_list = WorldList([world])

    # Run
    result_1 = list(world_list.connections_from(node_1, empty_patches))
    result_2 = list(world_list.connections_from(node_2, empty_patches))

    # Assert
    assert result_1 == [
        (node_2, RequirementAnd([req_1, req_2])),
    ]
    assert result_2 == [
        (node_1, req_2),
    ]



import pytest
from PySide2 import QtWidgets, QtCore
from mock import MagicMock, AsyncMock

from randovania.games.game import RandovaniaGame
from randovania.gui.generate_seed_tab import GenerateSeedTab
from randovania.gui.generated.main_window_ui import Ui_MainWindow
from randovania.interface_common.options import Options


@pytest.fixture(name="tab")
def _tab(skip_qtbot):
    class Window(QtWidgets.QMainWindow, Ui_MainWindow):
        pass

    window = Window()
    window.setupUi(window)

    return GenerateSeedTab(window, MagicMock(), MagicMock())


def test_add_new_preset(tab, preset_manager):
    preset = preset_manager.default_preset
    tab.setup_ui()
    tab.window.create_preset_tree.select_preset = MagicMock()

    # Run
    tab._add_new_preset(preset)

    # Assert
    tab._window_manager.preset_manager.add_new_preset.assert_called_once_with(preset)
    tab.window.create_preset_tree.select_preset.assert_called_once_with(preset)


@pytest.mark.parametrize("has_existing_window", [False, True])
@pytest.mark.asyncio
async def test_on_customize_button(tab, mocker, has_existing_window):
    mock_settings_window = mocker.patch("randovania.gui.generate_seed_tab.LogicSettingsWindow")
    mock_execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)
    mock_execute_dialog.return_value = QtWidgets.QDialog.Accepted
    tab._add_new_preset = MagicMock()
    tab._logic_settings_window = MagicMock() if has_existing_window else None
    tab.window.create_preset_tree = MagicMock()
    tab.window.create_preset_tree.selectedItems.return_value = [MagicMock()]

    # Run
    await tab._on_customize_preset()

    # Assert
    if has_existing_window:
        tab._logic_settings_window.raise_.assert_called_once_with()
        mock_settings_window.assert_not_called()
        mock_execute_dialog.assert_not_awaited()
        tab._add_new_preset.assert_not_called()
    else:
        mock_settings_window.assert_called_once()
        mock_execute_dialog.assert_awaited_once_with(mock_settings_window.return_value)
        tab._add_new_preset.assert_called_once()


@pytest.mark.parametrize("game", RandovaniaGame)
def test_on_options_changed_select_preset(tab, preset_manager, game: RandovaniaGame):
    preset = preset_manager.default_preset_for_game(game)

    tab._window_manager.preset_manager = preset_manager
    tab.setup_ui()

    tab._options.selected_preset_uuid = preset.uuid

    # Run
    tab.on_options_changed(tab._options)

    # Assert
    assert tab._current_preset_data == preset


@pytest.mark.parametrize("allow_experimental", [False, True])
@pytest.mark.parametrize("game", RandovaniaGame)
def test_click_on_preset_tree(tab, preset_manager, game: RandovaniaGame, skip_qtbot, tmp_path, allow_experimental):
    preset = preset_manager.default_preset_for_game(game)

    options = Options(tmp_path, None)
    with options:
        options.experimental_games = allow_experimental
    options.on_options_changed = lambda: tab.on_options_changed(options)

    tab._options = options
    tab._window_manager.preset_manager = preset_manager
    tab.setup_ui()
    tab.on_options_changed(options)

    # Run
    item = tab.window.create_preset_tree.preset_to_item.get(preset.uuid)
    # assert item.parent().text(0) == "1"
    if game.is_experimental and not allow_experimental:
        assert item is None
    else:
        tab.window.create_preset_tree.selectionModel().reset()
        tab.window.create_preset_tree.setItemSelected(item, True)

        # Assert
        assert tab._current_preset_data.get_preset() == preset.get_preset()



from PySide2 import QtCore

from randovania.gui.generated.preset_echoes_goal_ui import Ui_PresetEchoesGoal
from randovania.gui.lib.common_qt_lib import set_combo_with_value
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.prime2.echoes_configuration import LayoutSkyTempleKeyMode
from randovania.layout.preset import Preset


class PresetEchoesGoal(PresetTab, Ui_PresetEchoesGoal):

    def __init__(self, editor: PresetEditor):
        super().__init__(editor)
        self.setupUi(self)

        self.goal_layout.setAlignment(QtCore.Qt.AlignTop)
        self.skytemple_combo.setItemData(0, LayoutSkyTempleKeyMode.ALL_BOSSES)
        self.skytemple_combo.setItemData(1, LayoutSkyTempleKeyMode.ALL_GUARDIANS)
        self.skytemple_combo.setItemData(2, int)

        self.skytemple_combo.options_field_name = "layout_configuration_sky_temple_keys"
        self.skytemple_combo.currentIndexChanged.connect(self._on_sky_temple_key_combo_changed)
        self.skytemple_slider.valueChanged.connect(self._on_sky_temple_key_combo_slider_changed)

        # Default to False, since the slider defaults to a non-collect keys value.
        self._set_slider_visible(False)

    @property
    def uses_patches_tab(self) -> bool:
        return False

    def _set_slider_visible(self, visible: bool):
        for w in [self.skytemple_slider, self.skytemple_slider_label]:
            w.setVisible(visible)

    def _on_sky_temple_key_combo_changed(self):
        combo_enum = self.skytemple_combo.currentData()
        with self._editor as editor:
            if combo_enum is int:
                new_value = LayoutSkyTempleKeyMode(self.skytemple_slider.value())
                self._set_slider_visible(True)
            else:
                new_value = combo_enum
                self._set_slider_visible(False)

            editor.layout_configuration_sky_temple_keys = new_value

    def _on_sky_temple_key_combo_slider_changed(self):
        self.skytemple_slider_label.setText(str(self.skytemple_slider.value()))
        self._on_sky_temple_key_combo_changed()

    def on_preset_changed(self, preset: Preset):
        keys = preset.configuration.sky_temple_keys
        if isinstance(keys.value, int):
            self.skytemple_slider.setValue(keys.value)
            data = int
        else:
            data = keys
        set_combo_with_value(self.skytemple_combo, data)



from unittest.mock import MagicMock

import pytest

from randovania.game_description.resources.resource_type import ResourceType
from randovania.games.game import RandovaniaGame
from randovania.resolver import bootstrap


@pytest.mark.parametrize("vanilla_elevators", [False, True])
def test_misc_resources_for_configuration(echoes_resource_database,
                                          vanilla_elevators: bool,
                                          ):
    # Setup
    configuration = MagicMock()
    configuration.game = RandovaniaGame.METROID_PRIME_ECHOES
    configuration.elevators.is_vanilla = vanilla_elevators
    gfmc_resource = echoes_resource_database.get_by_type_and_index(ResourceType.MISC, 16)
    torvus_resource = echoes_resource_database.get_by_type_and_index(ResourceType.MISC, 17)
    great_resource = echoes_resource_database.get_by_type_and_index(ResourceType.MISC, 18)

    # Run
    result = bootstrap.misc_resources_for_configuration(configuration, echoes_resource_database)
    relevant_tricks = {
        trick: result[trick]
        for trick in [gfmc_resource, torvus_resource, great_resource]
    }

    # Assert
    assert relevant_tricks == {
        gfmc_resource: 0,
        torvus_resource: 0,
        great_resource: 0 if not vanilla_elevators else 1,
    }


def test_logic_bootstrap(default_preset, echoes_game_description):
    new_game, state = bootstrap.logic_bootstrap(default_preset.configuration,
                                                echoes_game_description.make_mutable_copy(),
                                                echoes_game_description.create_game_patches())


@pytest.mark.parametrize(["expected", "suits"], [
    (1.0, []),
    (0.9, ["Varia Suit"]),
    (0.9, ["Gravity Suit"]),
    (0.9, ["Phazon Suit"]),
    (0.8, ["Varia Suit", "Phazon Suit"]),
    (0.8, ["Varia Suit", "Gravity Suit"]),
    (0.8, ["Gravity Suit", "Phazon Suit"]),
    (0.5, ["Varia Suit", "Gravity Suit", "Phazon Suit"]),
])
def test_prime1_progressive_damage_reduction(prime1_resource_database, expected, suits):
    # Setup
    current_resources = {
        prime1_resource_database.get_item_by_name(suit): 1
        for suit in suits
    }

    # Run
    result = bootstrap.prime1_progressive_damage_reduction(prime1_resource_database, current_resources)

    # Assert
    assert result == expected


@pytest.mark.parametrize(["expected", "suits"], [
    (1.0, []),
    (0.9, ["Varia Suit"]),
    (0.8, ["Gravity Suit"]),
    (0.5, ["Phazon Suit"]),
    (0.5, ["Varia Suit", "Phazon Suit"]),
    (0.8, ["Varia Suit", "Gravity Suit"]),
    (0.5, ["Gravity Suit", "Phazon Suit"]),
    (0.5, ["Varia Suit", "Gravity Suit", "Phazon Suit"]),
])
def test_prime1_absolute_damage_reduction(prime1_resource_database, expected, suits):
    # Setup
    current_resources = {
        prime1_resource_database.get_item_by_name(suit): 1
        for suit in suits
    }

    # Run
    result = bootstrap.prime1_absolute_damage_reduction(prime1_resource_database, current_resources)

    # Assert
    assert result == expected



import dataclasses
from typing import List

import py_randomprime

from randovania.dol_patching import assembler
from randovania.dol_patching.assembler import custom_ppc
from randovania.dol_patching.assembler.ppc import lwz, r3, r13, bl, r31, r4, r5, cmpw, beq, li, stw, r1, addi, r6
from randovania.games.game import RandovaniaGame
from randovania.games.prime.all_prime_dol_patches import BasePrimeDolVersion, StringDisplayPatchAddresses, \
    PowerupFunctionsAddresses


@dataclasses.dataclass(frozen=True)
class Prime1DolVersion(BasePrimeDolVersion):
    state_for_world: int
    set_layer_active: int

    def __init__(self, version: str, description: str, build_string_address: int,
                 build_string: bytes, sda2_base: int, sda13_base: int, cplayer_vtable: int,
                 message_receiver_string_ref: int):
        symbols = py_randomprime.symbols_for_version(version)

        super().__init__(
            game=RandovaniaGame.METROID_PRIME,
            description=description,
            build_string_address=build_string_address,
            build_string=build_string,
            sda2_base=sda2_base,
            sda13_base=sda13_base,
            game_state_pointer=symbols["g_GameState"],
            cplayer_vtable=cplayer_vtable,
            cstate_manager_global=symbols["g_StateManager"],
            string_display=StringDisplayPatchAddresses(
                update_hint_state=symbols["UpdateHintState__13CStateManagerFf"],
                message_receiver_string_ref=message_receiver_string_ref,
                wstring_constructor=symbols["wstring_l__4rstlFPCw"],
                display_hud_memo=symbols["DisplayHudMemo__9CSamusHudFRC7wstringRC12SHudMemoInfo"],
                max_message_size=200,
            ),
            powerup_functions=PowerupFunctionsAddresses(
                add_power_up=symbols["InitializePowerUp__12CPlayerStateFQ212CPlayerState9EItemTypei"],
                incr_pickup=symbols["IncrPickUp__12CPlayerStateFQ212CPlayerState9EItemTypei"],
                decr_pickup=symbols["DecrPickUp__12CPlayerStateFQ212CPlayerState9EItemTypei"],
            ),
        )
        object.__setattr__(self, "state_for_world", symbols["StateForWorld__10CGameStateFUi"])
        object.__setattr__(self, "set_layer_active", symbols["SetLayerActive__16CWorldLayerStateFiib"])


def set_artifact_layer_active_patch(addresses: Prime1DolVersion, layer_id: int, active: bool,
                                    ) -> List[assembler.BaseInstruction]:
    # g_GameState->StateForWorld(0x39F2DE28)->GetLayerState()->SetLayerActive(templeAreaIndex, artifactLayer, true)
    result = []

    for_another_world = [
        # Get the LayerState via the CGameState
        lwz(r3, addresses.game_state_pointer - addresses.sda13_base, r13),  # get g_GameState
        # r4 already have the asset id
        bl(addresses.state_for_world),  # CGameState::StateForWorld
        lwz(r3, 0x14, r3),  # worldState->layerState
    ]

    result.extend([
        # Get the LayerState of current world. We'll overwrite if it's another world, it's just 1 instruction bigger
        lwz(r3, 0x8c8, r31),  # mgr->worldLayerState

        # Tallon Overworld asset id
        custom_ppc.load_unsigned_32bit(r4, 0x39f2de28),

        # Load current asset id in r5
        lwz(r5, 0x850, r31),  # mgr->world
        lwz(r5, 0x8, r5),  # world->mlvlId

        cmpw(0, r4, r5),  # compare asset ids
        beq(4 + assembler.byte_count(for_another_world), relative=True),
        *for_another_world,
        lwz(r3, 0x0, r3),

        # Set layer
        li(r4, 16),  # Artifact Layer
        stw(r4, 0x10, r1),

        # Set layer
        li(r5, layer_id),  # Artifact Layer
        stw(r5, 0x14, r1),

        # Make the layer change via SetLayerActive
        addi(r4, r1, 0x10),
        addi(r5, r1, 0x14),
        li(r6, int(active)),
        bl(addresses.set_layer_active),  # CWorldLayerState::SetLayerActive
    ])

    return result



import dataclasses

from randovania.games.game import RandovaniaGame
from randovania.layout.base.cosmetic_patches import BaseCosmeticPatches
from randovania.layout.prime2.echoes_user_preferences import EchoesUserPreferences


@dataclasses.dataclass(frozen=True)
class EchoesCosmeticPatches(BaseCosmeticPatches):
    disable_hud_popup: bool = True
    speed_up_credits: bool = True
    open_map: bool = True
    unvisited_room_names: bool = True
    pickup_markers: bool = True
    teleporter_sounds: bool = True
    user_preferences: EchoesUserPreferences = dataclasses.field(default_factory=EchoesUserPreferences)
    convert_other_game_assets: bool = False

    @classmethod
    def default(cls) -> "EchoesCosmeticPatches":
        return cls()

    @classmethod
    def game(cls):
        return RandovaniaGame.METROID_PRIME_ECHOES



import dataclasses

from randovania.games.game import RandovaniaGame
from randovania.layout.base.base_configuration import BaseConfiguration


@dataclasses.dataclass(frozen=True)
class SuperMetroidConfiguration(BaseConfiguration):

    @classmethod
    def game_enum(cls) -> RandovaniaGame:
        return RandovaniaGame.SUPER_METROID



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_echoes_goal.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetEchoesGoal(object):
    def setupUi(self, PresetEchoesGoal):
        if not PresetEchoesGoal.objectName():
            PresetEchoesGoal.setObjectName(u"PresetEchoesGoal")
        PresetEchoesGoal.resize(562, 459)
        self.centralWidget = QWidget(PresetEchoesGoal)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.goal_layout = QVBoxLayout(self.centralWidget)
        self.goal_layout.setSpacing(6)
        self.goal_layout.setContentsMargins(11, 11, 11, 11)
        self.goal_layout.setObjectName(u"goal_layout")
        self.goal_layout.setContentsMargins(4, 8, 4, 8)
        self.skytemple_description = QLabel(self.centralWidget)
        self.skytemple_description.setObjectName(u"skytemple_description")
        self.skytemple_description.setWordWrap(True)

        self.goal_layout.addWidget(self.skytemple_description)

        self.skytemple_combo = QComboBox(self.centralWidget)
        self.skytemple_combo.addItem("")
        self.skytemple_combo.addItem("")
        self.skytemple_combo.addItem("")
        self.skytemple_combo.setObjectName(u"skytemple_combo")

        self.goal_layout.addWidget(self.skytemple_combo)

        self.skytemple_slider_layout = QHBoxLayout()
        self.skytemple_slider_layout.setSpacing(6)
        self.skytemple_slider_layout.setObjectName(u"skytemple_slider_layout")
        self.skytemple_slider = QSlider(self.centralWidget)
        self.skytemple_slider.setObjectName(u"skytemple_slider")
        self.skytemple_slider.setMaximum(9)
        self.skytemple_slider.setPageStep(2)
        self.skytemple_slider.setOrientation(Qt.Horizontal)
        self.skytemple_slider.setTickPosition(QSlider.TicksBelow)

        self.skytemple_slider_layout.addWidget(self.skytemple_slider)

        self.skytemple_slider_label = QLabel(self.centralWidget)
        self.skytemple_slider_label.setObjectName(u"skytemple_slider_label")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.skytemple_slider_label.sizePolicy().hasHeightForWidth())
        self.skytemple_slider_label.setSizePolicy(sizePolicy)
        self.skytemple_slider_label.setMinimumSize(QSize(20, 0))
        self.skytemple_slider_label.setAlignment(Qt.AlignCenter)

        self.skytemple_slider_layout.addWidget(self.skytemple_slider_label)


        self.goal_layout.addLayout(self.skytemple_slider_layout)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.goal_layout.addItem(self.horizontalSpacer)

        PresetEchoesGoal.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetEchoesGoal)

        QMetaObject.connectSlotsByName(PresetEchoesGoal)
    # setupUi

    def retranslateUi(self, PresetEchoesGoal):
        PresetEchoesGoal.setWindowTitle(QCoreApplication.translate("PresetEchoesGoal", u"Goal", None))
        self.skytemple_description.setText(QCoreApplication.translate("PresetEchoesGoal", u"<html><head/><body><p>Controls where the Sky Temple Keys will be located.</p><p>All Guardians and Sub-Guardians: One key will be placed in each of the guardians and sub-guardians.<br/>Guardians: One key will be placed as the reward of each of the guardians.<br/>Collect Sky Temple Keys: A configurable quantity will be shuffled over the game world.</p><p>The Guardians are: Amorbis, Chykka and Quadraxis.<br/>The Sub-Guardians are: Bomb Guardian, Jump Guardian, Boost Guardian, Grapple Guardian, Spider Guardian and Power Bomb Guardian.</p><p>You can always check Sky Temple Gateway for hints where the keys were placed.</p></body></html>", None))
        self.skytemple_combo.setItemText(0, QCoreApplication.translate("PresetEchoesGoal", u"Guardians and Sub-Guardians", None))
        self.skytemple_combo.setItemText(1, QCoreApplication.translate("PresetEchoesGoal", u"Guardians", None))
        self.skytemple_combo.setItemText(2, QCoreApplication.translate("PresetEchoesGoal", u"Collect Sky Temple Keys", None))

        self.skytemple_slider_label.setText(QCoreApplication.translate("PresetEchoesGoal", u"0", None))
    # retranslateUi




import os
import platform
import subprocess
import typing
from pathlib import Path
from typing import Iterator, Optional

from PySide2 import QtWidgets, QtGui

from randovania import get_data_path


def map_set_checked(iterable: Iterator[QtWidgets.QCheckBox], new_status: bool):
    for checkbox in iterable:
        checkbox.setChecked(new_status)


def lock_application(value: bool):
    QtWidgets.QApplication.instance().main_window.setEnabled(value)


def _prompt_user_for_file(window: QtWidgets.QMainWindow,
                          caption: str,
                          filter: str,
                          dir: Optional[str] = None,
                          new_file: bool = False) -> Optional[Path]:
    """
    Helper function for all `prompt_user_for_*` functions.
    :param window:
    :param caption:
    :param filter:
    :param new_file: If false, prompt for an existing file.
    :return: A string if the user selected a file, None otherwise
    """
    if new_file:
        method = QtWidgets.QFileDialog.getSaveFileName
    else:
        method = QtWidgets.QFileDialog.getOpenFileName
    open_result = method(window, caption=caption, dir=dir, filter=filter)
    if not open_result or open_result == ("", ""):
        return None
    return Path(open_result[0])


def prompt_user_for_vanilla_input_file(window: QtWidgets.QMainWindow, extensions: typing.List[str],
                                       existing_file: Optional[Path] = None) -> Optional[Path]:
    """
    Shows an QFileDialog asking the user for a vanilla game file
    :param window:
    :param extensions:
    :param existing_file: An existing file to pre-fill with.
    :return: A string if the user selected a file, None otherwise
    """
    return _prompt_user_for_file(
        window,
        caption="Select the vanilla game {}.".format("/".join(extensions)),
        dir=str(existing_file) if existing_file is not None else None,
        filter=";".join(f"*.{ext}" for ext in extensions),
    )


def prompt_user_for_output_file(window: QtWidgets.QMainWindow,
                                default_name: str,
                                extensions: typing.List[str]) -> Optional[Path]:
    """
    Shows an QFileDialog asking the user where to place the output file
    :param window:
    :param default_name: Name of a file that will be offered by default in the UI.
    :param extensions:
    :return: A string if the user selected a file, None otherwise
    """
    return _prompt_user_for_file(
        window,
        caption="Where to place the Randomized game file.",
        dir=default_name,
        filter=";".join(f"*.{ext}" for ext in extensions),
        new_file=True,
    )


def prompt_user_for_output_game_log(window: QtWidgets.QMainWindow, default_name: str) -> Optional[Path]:
    """
    Shows an QFileDialog asking the user for a Randovania seed log
    :param window:
    :param default_name:
    :return: A string if the user selected a file, None otherwise
    """
    from randovania.layout.layout_description import LayoutDescription
    return _prompt_user_for_file(window, caption="Select a Randovania seed log.",
                                 dir=default_name,
                                 filter="Randovania Game, *.{}".format(LayoutDescription.file_extension()),
                                 new_file=True)


def prompt_user_for_input_game_log(window: QtWidgets.QMainWindow) -> Optional[Path]:
    """
    Shows an QFileDialog asking the user for a Randovania seed log
    :param window:
    :return: A string if the user selected a file, None otherwise
    """
    from randovania.layout.layout_description import LayoutDescription
    return _prompt_user_for_file(window, caption="Select a Randovania seed log.",
                                 filter="Randovania Game, *.{}".format(LayoutDescription.file_extension()),
                                 new_file=False)


def prompt_user_for_database_file(window: QtWidgets.QMainWindow) -> Optional[Path]:
    """
    Shows an QFileDialog asking the user for a Randovania database file
    :param window:
    :return: A string if the user selected a file, None otherwise
    """
    return _prompt_user_for_file(window, caption="Select a Randovania database file.", filter="*.json")


def prompt_user_for_preset_file(window: QtWidgets.QMainWindow, new_file: bool) -> Optional[Path]:
    """
    Shows an QFileDialog asking the user for a Randovania preset file
    :param window:
    :param new_file: If it should be an existing file (False) or not.
    :return: A path if the user selected a file, None otherwise
    """
    from randovania.layout.preset_migration import VersionedPreset
    return _prompt_user_for_file(window, caption="Select a Randovania Preset file.",
                                 filter=f"Randovania Preset, *.{VersionedPreset.file_extension()};;"
                                        f"All Files (*.*)",
                                 new_file=new_file)


def set_default_window_icon(window: QtWidgets.QWidget):
    """
    Sets the window icon for the given widget to the default icon
    :param window:
    :return:
    """
    window.setWindowIcon(QtGui.QIcon(str(get_data_path().joinpath("icons", "sky_temple_key_NqN_icon.ico"))))


def set_combo_with_value(combo: QtWidgets.QComboBox, value):
    """
    Searches all items of the given combo for the given value and changes the current index to that one.
    :param combo:
    :param value:
    :return:
    """
    combo.setCurrentIndex(combo.findData(value))


def set_error_border_stylesheet(edit: QtWidgets.QTextEdit, has_error: bool):
    edit.has_error = has_error
    if has_error:
        edit.setStyleSheet(":enabled { border: 1px solid red; }"
                           ":disabled { border: 1px solid red; background: #CCC }")
    else:
        edit.setStyleSheet("")


def set_edit_if_different(edit: QtWidgets.QLineEdit, new_text: str):
    """
    Sets the text of the given QLineEdit only if it differs from the current value.
    Prevents snapping the user's cursor to the end unnecessarily.
    :param edit:
    :param new_text:
    :return:
    """
    if edit.text() != new_text:
        edit.setText(new_text)


def get_network_client():
    from randovania.gui.lib.qt_network_client import QtNetworkClient
    return typing.cast(QtNetworkClient, QtWidgets.QApplication.instance().network_client)


def get_game_connection():
    from randovania.game_connection.game_connection import GameConnection
    return typing.cast(GameConnection, QtWidgets.QApplication.instance().game_connection)


def show_install_visual_cpp_redist(details: str):
    from PySide2 import QtWidgets

    download_url = 'https://aka.ms/vs/16/release/vc_redist.x64.exe'
    support_url = 'https://support.microsoft.com/en-us/help/2977003/the-latest-supported-visual-c-downloads'

    box = QtWidgets.QMessageBox(
        QtWidgets.QMessageBox.Critical,
        "Unable to load Dolphin backend",
        "Please install the latest "
        f"<a href='{download_url}'>Microsoft Visual C++ Redistributable</a>.<br /><br />"
        f"For more details, see <a href='{support_url}'>Microsoft's webpage</a>.",
        QtWidgets.QMessageBox.Ok,
    )
    set_default_window_icon(box)
    box.setDetailedText(details)
    box.exec_()


def set_clipboard(text: str):
    from PySide2 import QtWidgets
    QtWidgets.QApplication.clipboard().setText(text)


def open_directory_in_explorer(path: Path):
    if platform.system() == "Windows":
        os.startfile(path)
    elif platform.system() == "Darwin":
        subprocess.run(["open", path])
    else:
        subprocess.run(["xdg-open", path])



import json

import pytest

from randovania.game_description import data_reader, data_writer
from randovania.games import default_data
from randovania.games.game import RandovaniaGame


@pytest.mark.parametrize("game_enum", RandovaniaGame)
def test_round_trip_full(game_enum: RandovaniaGame):
    original_data = default_data.read_json_then_binary(game_enum)[1]

    game = data_reader.decode_data(original_data)
    encoded_data = data_writer.write_game_description(game)

    assert list(encoded_data.keys()) == list(original_data.keys())
    assert encoded_data == original_data


@pytest.mark.parametrize("small_name", ["prime_data_as_json.json", "prime2_small.json"])
def test_round_trip_small(test_files_dir, small_name):
    # Setup
    with test_files_dir.joinpath(small_name).open("r") as data_file:
        original_data = json.load(data_file)

    game = data_reader.decode_data(original_data)
    encoded_data = data_writer.write_game_description(game)

    # # Uncomment the following to update the file
    # with test_files_dir.joinpath(small_name).open("w", encoding="utf-8") as meta:
    #     json.dump(encoded_data, meta, indent=4); assert False

    assert encoded_data == original_data



import functools
from typing import List, Callable, FrozenSet, Dict

from PySide2 import QtWidgets, QtCore

from randovania.games.prime import elevators
from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.game_description import GameDescription
from randovania.game_description.world.world import World


def dark_world_flags(world: World):
    yield False
    if world.dark_name is not None:
        yield True


class AreaListHelper:
    game_description: GameDescription
    during_batch_check_update: bool

    def areas_by_world_from_locations(self, all_area_locations: List[AreaLocation]):
        worlds = []
        areas_by_world = {}

        for location in all_area_locations:
            world = self.game_description.world_list.world_by_asset_id(location.world_asset_id)
            if world.world_asset_id not in areas_by_world:
                worlds.append(world)
                areas_by_world[world.world_asset_id] = []

            areas_by_world[world.world_asset_id].append(world.area_by_asset_id(location.area_asset_id))

        return worlds, areas_by_world

    def create_area_list_selection(self, parent: QtWidgets.QWidget, layout: QtWidgets.QGridLayout,
                                   all_area_locations: List[AreaLocation],
                                   on_check: Callable[[List[AreaLocation], bool], None],
                                   ):
        world_to_group = {}
        checks_for_world = {}
        checks_for_area = {}

        worlds, areas_by_world = self.areas_by_world_from_locations(all_area_locations)
        worlds.sort(key=lambda it: it.name)

        def _on_check_area(c, _):
            if not self.during_batch_check_update:
                on_check([c.area_location], c.isChecked())

        def _on_check_world(c, _):
            if not self.during_batch_check_update:
                world_list = self.game_description.world_list
                w = world_list.world_by_asset_id(c.world_asset_id)
                world_areas = [world_list.area_to_area_location(a)
                               for a in w.areas if c.is_dark_world == a.in_dark_aether
                               if a.area_asset_id in checks_for_area]
                on_check(world_areas, c.isChecked())

        for row, world in enumerate(worlds):
            for column, is_dark_world in enumerate(dark_world_flags(world)):
                group_box = QtWidgets.QGroupBox(parent)
                world_check = QtWidgets.QCheckBox(group_box)
                world_check.setText(world.correct_name(is_dark_world))
                world_check.world_asset_id = world.world_asset_id
                world_check.is_dark_world = is_dark_world
                world_check.stateChanged.connect(functools.partial(_on_check_world, world_check))
                world_check.setTristate(True)
                vertical_layout = QtWidgets.QVBoxLayout(group_box)
                vertical_layout.setContentsMargins(8, 4, 8, 4)
                vertical_layout.setSpacing(2)
                vertical_layout.setAlignment(QtCore.Qt.AlignTop)
                separator = QtWidgets.QFrame()
                separator.setFrameShape(QtWidgets.QFrame.HLine)
                separator.setFrameShadow(QtWidgets.QFrame.Sunken)
                group_box.vertical_layout = vertical_layout
                group_box.vertical_layout.addWidget(world_check)
                group_box.vertical_layout.addWidget(separator)

                world_to_group[world.correct_name(is_dark_world)] = group_box
                layout.addWidget(group_box, row, column)
                checks_for_world[world.correct_name(is_dark_world)] = world_check

        for world in worlds:
            for area in sorted(areas_by_world[world.world_asset_id], key=lambda a: a.name):
                group_box = world_to_group[world.correct_name(area.in_dark_aether)]
                check = QtWidgets.QCheckBox(group_box)
                check.setText(elevators.get_elevator_name_or_default(self.game_description.game, area.area_asset_id, area.name))
                check.area_location = AreaLocation(world.world_asset_id, area.area_asset_id)
                check.stateChanged.connect(functools.partial(_on_check_area, check))
                group_box.vertical_layout.addWidget(check)
                checks_for_area[area.area_asset_id] = check

        return checks_for_world, checks_for_area

    def update_area_list(self, areas_to_check: FrozenSet[AreaLocation],
                         invert_check: bool,
                         location_for_world: Dict[str, QtWidgets.QCheckBox],
                         location_for_area: Dict[int, QtWidgets.QCheckBox],
                         ):
        self.during_batch_check_update = True

        for world in self.game_description.world_list.worlds:
            for is_dark_world in dark_world_flags(world):
                all_areas = True
                no_areas = True
                areas = [area for area in world.areas if area.in_dark_aether == is_dark_world]
                correct_name = world.correct_name(is_dark_world)
                if correct_name not in location_for_world:
                    continue

                for area in areas:
                    if area.area_asset_id in location_for_area:
                        is_checked = AreaLocation(world.world_asset_id, area.area_asset_id) in areas_to_check
                        if invert_check:
                            is_checked = not is_checked

                        if is_checked:
                            no_areas = False
                        else:
                            all_areas = False
                        location_for_area[area.area_asset_id].setChecked(is_checked)
                if all_areas:
                    location_for_world[correct_name].setCheckState(QtCore.Qt.Checked)
                elif no_areas:
                    location_for_world[correct_name].setCheckState(QtCore.Qt.Unchecked)
                else:
                    location_for_world[correct_name].setCheckState(QtCore.Qt.PartiallyChecked)

        self.during_batch_check_update = False



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'widget_location_pool_row.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_LocationPoolRowWidget(object):
    def setupUi(self, LocationPoolRowWidget):
        if not LocationPoolRowWidget.objectName():
            LocationPoolRowWidget.setObjectName(u"LocationPoolRowWidget")
        LocationPoolRowWidget.resize(616, 52)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(LocationPoolRowWidget.sizePolicy().hasHeightForWidth())
        LocationPoolRowWidget.setSizePolicy(sizePolicy)
        LocationPoolRowWidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(LocationPoolRowWidget)
        self.verticalLayout.setSpacing(4)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 2, 0, 0)
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setSpacing(6)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.label_location_name = QLabel(LocationPoolRowWidget)
        self.label_location_name.setObjectName(u"label_location_name")
        font = QFont()
        font.setPointSize(8)
        font.setBold(True)
        font.setWeight(75)
        self.label_location_name.setFont(font)

        self.horizontalLayout_2.addWidget(self.label_location_name)

        self.radio_shuffled = QRadioButton(LocationPoolRowWidget)
        self.radio_shuffled.setObjectName(u"radio_shuffled")
        sizePolicy1 = QSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.radio_shuffled.sizePolicy().hasHeightForWidth())
        self.radio_shuffled.setSizePolicy(sizePolicy1)
        self.radio_shuffled.setChecked(True)

        self.horizontalLayout_2.addWidget(self.radio_shuffled)

        self.radio_shuffled_no_progression = QRadioButton(LocationPoolRowWidget)
        self.radio_shuffled_no_progression.setObjectName(u"radio_shuffled_no_progression")
        sizePolicy1.setHeightForWidth(self.radio_shuffled_no_progression.sizePolicy().hasHeightForWidth())
        self.radio_shuffled_no_progression.setSizePolicy(sizePolicy1)

        self.horizontalLayout_2.addWidget(self.radio_shuffled_no_progression)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(12)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setSizeConstraint(QLayout.SetDefaultConstraint)
        self.horizontalLayout.setContentsMargins(-1, 0, -1, -1)
        self.radio_vanilla = QRadioButton(LocationPoolRowWidget)
        self.radio_vanilla.setObjectName(u"radio_vanilla")
        sizePolicy2 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.radio_vanilla.sizePolicy().hasHeightForWidth())
        self.radio_vanilla.setSizePolicy(sizePolicy2)

        self.horizontalLayout.addWidget(self.radio_vanilla)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setSpacing(0)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.radio_fixed = QRadioButton(LocationPoolRowWidget)
        self.radio_fixed.setObjectName(u"radio_fixed")
        sizePolicy1.setHeightForWidth(self.radio_fixed.sizePolicy().hasHeightForWidth())
        self.radio_fixed.setSizePolicy(sizePolicy1)

        self.horizontalLayout_3.addWidget(self.radio_fixed)

        self.combo_fixed_item = QComboBox(LocationPoolRowWidget)
        self.combo_fixed_item.addItem("")
        self.combo_fixed_item.setObjectName(u"combo_fixed_item")
        sizePolicy3 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.combo_fixed_item.sizePolicy().hasHeightForWidth())
        self.combo_fixed_item.setSizePolicy(sizePolicy3)

        self.horizontalLayout_3.addWidget(self.combo_fixed_item)


        self.horizontalLayout.addLayout(self.horizontalLayout_3)


        self.verticalLayout.addLayout(self.horizontalLayout)


        self.retranslateUi(LocationPoolRowWidget)

        QMetaObject.connectSlotsByName(LocationPoolRowWidget)
    # setupUi

    def retranslateUi(self, LocationPoolRowWidget):
        LocationPoolRowWidget.setWindowTitle(QCoreApplication.translate("LocationPoolRowWidget", u"\n"
"				Item Configuration\n"
"			", None))
        self.label_location_name.setText(QCoreApplication.translate("LocationPoolRowWidget", u"Location name", None))
#if QT_CONFIG(tooltip)
        self.radio_shuffled.setToolTip(QCoreApplication.translate("LocationPoolRowWidget", u"This location is shuffled normally", None))
#endif // QT_CONFIG(tooltip)
        self.radio_shuffled.setText(QCoreApplication.translate("LocationPoolRowWidget", u"Shuffled", None))
#if QT_CONFIG(tooltip)
        self.radio_shuffled_no_progression.setToolTip(QCoreApplication.translate("LocationPoolRowWidget", u"This location is shuffled, but cannot contain a item required for progression", None))
#endif // QT_CONFIG(tooltip)
        self.radio_shuffled_no_progression.setText(QCoreApplication.translate("LocationPoolRowWidget", u"Shuffled (no progression)", None))
#if QT_CONFIG(tooltip)
        self.radio_vanilla.setToolTip(QCoreApplication.translate("LocationPoolRowWidget", u"This location is unchanged and will contain the same item as in the base game", None))
#endif // QT_CONFIG(tooltip)
        self.radio_vanilla.setText(QCoreApplication.translate("LocationPoolRowWidget", u"Vanilla", None))
#if QT_CONFIG(tooltip)
        self.radio_fixed.setToolTip(QCoreApplication.translate("LocationPoolRowWidget", u"This location contains a fixed item of your choosing", None))
#endif // QT_CONFIG(tooltip)
        self.radio_fixed.setText(QCoreApplication.translate("LocationPoolRowWidget", u"Fixed			", None))
        self.combo_fixed_item.setItemText(0, QCoreApplication.translate("LocationPoolRowWidget", u"No item", None))

        self.combo_fixed_item.setCurrentText(QCoreApplication.translate("LocationPoolRowWidget", u"No item", None))
    # retranslateUi




import io
import json
import logging
import re

import discord
from discord.ext import commands
from discord_slash import ComponentContext, ButtonStyle
from discord_slash.utils import manage_components

import randovania
from randovania.gui.lib import preset_describer
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.permalink import Permalink
from randovania.layout.preset import Preset
from randovania.layout.preset_migration import VersionedPreset
from randovania.server.discord.bot import RandovaniaBot

possible_links_re = re.compile(r'([A-Za-z0-9-_]{8,})')


def _add_preset_description_to_embed(embed: discord.Embed, preset: Preset):
    for category, items in preset_describer.describe(preset):
        embed.add_field(name=category, value="\n".join(items), inline=True)


async def look_for_permalinks(message: discord.Message):
    embed = None
    multiple_permalinks = False
    components = []

    for word in possible_links_re.finditer(message.content):
        try:
            permalink = Permalink.from_str(word.group(1))
        except ValueError:
            continue

        if embed is not None:
            multiple_permalinks = True
            continue

        embed = discord.Embed(title=f"`{word.group(1)}`",
                              description="{} player multiworld permalink".format(permalink.player_count))

        if permalink.player_count == 1:
            preset = permalink.get_preset(0)
            embed.description = "{} permalink for Randovania {}".format(preset.game.long_name,
                                                                        randovania.VERSION)
            _add_preset_description_to_embed(embed, preset)

        components.append(manage_components.create_actionrow(manage_components.create_button(
            style=ButtonStyle.secondary,
            label="Attach presets",
            custom_id="attach_presets_of_permalink",
        )))

    if embed is not None:
        content = None
        if multiple_permalinks:
            content = "Multiple permalinks found, using only the first."
        await message.reply(content=content, embed=embed, components=components, mention_author=False)


async def reply_for_preset(message: discord.Message, versioned_preset: VersionedPreset):
    try:
        preset = versioned_preset.get_preset()
    except ValueError as e:
        logging.info("Invalid preset '{}' from {}: {}".format(versioned_preset.name,
                                                              message.author.display_name,
                                                              e))
        return

    embed = discord.Embed(title=preset.name,
                          description=preset.description)
    _add_preset_description_to_embed(embed, preset)
    await message.reply(embed=embed, mention_author=False)


async def reply_for_layout_description(message: discord.Message, description: LayoutDescription):
    embed = discord.Embed(
        title="Spoiler file for Randovania {}".format(description.version),
    )

    if description.permalink.player_count == 1:
        preset = description.get_preset(0)
        embed.description = "{}, with preset {}".format(preset.game.long_name, preset.name)
        _add_preset_description_to_embed(embed, preset)
    else:
        games = {preset.game.long_name for preset in description.permalink.presets.values()}
        game_names = sorted(games)

        last_game = game_names.pop()
        games_text = ", ".join(game_names)
        if games_text:
            games_text += " and "
        games_text += last_game

        embed.description = "{} player multiworld for {}.\nSeed Hash: {}\nPermalink: {}".format(
            description.permalink.player_count,
            games_text,
            description.shareable_word_hash,
            description.permalink.as_base64_str,
        )

    await message.reply(embed=embed, mention_author=False)


class PermalinkLookupCog(commands.Cog):
    def __init__(self, configuration: dict, bot: RandovaniaBot):
        self.configuration = configuration
        self.bot = bot

    @commands.Cog.listener()
    async def on_ready(self):
        self.bot.slash.add_component_callback(
            self.on_request_presets,
            components=["attach_presets_of_permalink"],
            use_callback_name=False,
        )

    @commands.Cog.listener()
    async def on_message(self, message: discord.Message):
        if message.author == self.bot.user:
            return

        if message.guild is not None and message.guild.id != self.configuration["guild"]:
            return

        for attachment in message.attachments:
            filename: str = attachment.filename
            if filename.endswith(VersionedPreset.file_extension()):
                data = await attachment.read()
                versioned_preset = VersionedPreset(json.loads(data.decode("utf-8")))
                await reply_for_preset(message, versioned_preset)

            elif filename.endswith(LayoutDescription.file_extension()):
                data = await attachment.read()
                description = LayoutDescription.from_json_dict(json.loads(data.decode("utf-8")))
                await reply_for_layout_description(message, description)

        await look_for_permalinks(message)

    async def on_request_presets(self, ctx: ComponentContext):
        try:
            title = ctx.origin_message.embeds[0].title
            # Trim leading and trailing `s
            permalink = Permalink.from_str(title[1:-1])

        except (IndexError, ValueError) as e:
            logging.exception("Unable to find permalink on message that sent attach_presets_of_permalink")
            permalink = None

        files = []

        if permalink is not None:
            for player, preset in permalink.presets.items():
                data = io.BytesIO()
                VersionedPreset.with_preset(preset).save_to_io(data)
                data.seek(0)
                files.append(
                    discord.File(data, filename=f"Player {player + 1}'s Preset.{VersionedPreset.file_extension()}")
                )

        await ctx.edit_origin(
            components=[],
            files=files,
        )


def setup(bot: RandovaniaBot):
    bot.add_cog(PermalinkLookupCog(bot.configuration, bot))



import randovania

SERVER_API_VERSION = 4


def connection_headers():
    from randovania.layout.preset_migration import CURRENT_PRESET_VERSION
    from randovania.layout.permalink import Permalink
    from randovania.layout.layout_description import CURRENT_DESCRIPTION_SCHEMA_VERSION
    return {
        "X-Randovania-Version": randovania.VERSION,
        "X-Randovania-API-Version": str(SERVER_API_VERSION),
        "X-Randovania-Preset-Version": str(CURRENT_PRESET_VERSION),
        "X-Randovania-Permalink-Version": str(Permalink.current_version()),
        "X-Randovania-Description-Version": str(CURRENT_DESCRIPTION_SCHEMA_VERSION),
    }



import collections
import copy
import dataclasses
from functools import partial
from typing import List, Dict, Optional, Set

from PySide2 import QtCore, QtWidgets, QtGui
from PySide2.QtCore import Qt
from PySide2.QtWidgets import QRadioButton, QGroupBox, QHBoxLayout, QLabel, QPushButton, QVBoxLayout, \
    QApplication, QDialog, QAction, QMenu
from qasync import asyncSlot

from randovania.game_description import default_database
from randovania.game_description.game_description import GameDescription
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.world.node import PickupNode
from randovania.games.game import RandovaniaGame
from randovania.games.prime.patcher_file_lib import item_names
from randovania.gui import game_specific_gui
from randovania.gui.dialog.game_input_dialog import GameInputDialog
from randovania.gui.dialog.scroll_label_dialog import ScrollLabelDialog
from randovania.gui.generated.seed_details_window_ui import Ui_SeedDetailsWindow
from randovania.gui.lib import preset_describer, async_dialog, common_qt_lib, game_exporter
from randovania.gui.lib.background_task_mixin import BackgroundTaskMixin
from randovania.gui.lib.close_event_widget import CloseEventWidget
from randovania.gui.lib.common_qt_lib import set_default_window_icon, prompt_user_for_output_game_log
from randovania.gui.lib.window_manager import WindowManager
from randovania.interface_common import simplified_patcher
from randovania.interface_common.options import Options, InfoAlert
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.preset_migration import VersionedPreset


def _unique(iterable):
    seen = set()

    for item in iterable:
        if item in seen:
            continue
        seen.add(item)
        yield item


def _show_pickup_spoiler(button):
    target_player = getattr(button, "target_player", None)
    if target_player is not None:
        label = f"{button.item_name} for {button.player_names[target_player]}"
    else:
        label = button.item_name
    button.setText(label)
    button.item_is_hidden = False


def _hide_pickup_spoiler(button):
    button.setText("Hidden")
    button.item_is_hidden = True


class SeedDetailsWindow(CloseEventWidget, Ui_SeedDetailsWindow, BackgroundTaskMixin):
    _on_bulk_change: bool = False
    _history_items: List[QRadioButton]
    pickup_spoiler_buttons: List[QPushButton]
    layout_description: LayoutDescription
    _options: Options
    _window_manager: Optional[WindowManager]
    _player_names: Dict[int, str]
    _pickup_spoiler_current_game: Optional[GameDescription] = None
    _last_percentage: float = 0
    _can_stop_background_process: bool = True

    def __init__(self, window_manager: Optional[WindowManager], options: Options):
        super().__init__()
        self.setupUi(self)
        set_default_window_icon(self)

        self._history_items = []
        self.pickup_spoiler_buttons = []
        self._pickup_spoiler_world_to_group = {}
        self._options = options
        self._window_manager = window_manager

        # Ui
        self._tool_button_menu = QMenu(self.tool_button)
        self.tool_button.setMenu(self._tool_button_menu)

        self._action_open_tracker = QAction(self)
        self._action_open_tracker.setText("Open map tracker")
        self._action_open_tracker.setEnabled(self._window_manager is not None)
        self._tool_button_menu.addAction(self._action_open_tracker)

        self._action_copy_permalink = QAction(self)
        self._action_copy_permalink.setText("Copy Permalink")
        self._tool_button_menu.addAction(self._action_copy_permalink)

        self._action_export_preset = QAction(self)
        self._action_export_preset.setText("Export current player's preset")
        self._tool_button_menu.addAction(self._action_export_preset)

        self._action_view_trick_usages = QAction(self)
        self._action_view_trick_usages.setText("View current player's expected trick usage")
        self._tool_button_menu.addAction(self._action_view_trick_usages)

        # Signals
        self.export_log_button.clicked.connect(self._export_log)
        self.export_iso_button.clicked.connect(self._export_iso)
        self._action_open_tracker.triggered.connect(self._open_map_tracker)
        self._action_copy_permalink.triggered.connect(self._copy_permalink)
        self._action_export_preset.triggered.connect(self._export_preset)
        self._action_view_trick_usages.triggered.connect(self._view_trick_usages)
        self.pickup_spoiler_pickup_combobox.currentTextChanged.connect(self._on_change_pickup_filter)
        self.pickup_spoiler_show_all_button.clicked.connect(self._toggle_show_all_pickup_spoiler)
        self.player_index_combo.activated.connect(self._update_current_player)
        self.CloseEvent.connect(self.stop_background_process)

        # Progress
        self.background_tasks_button_lock_signal.connect(self.enable_buttons_with_background_tasks)
        self.progress_update_signal.connect(self.update_progress)
        self.stop_background_process_button.clicked.connect(self.stop_background_process)

        # Cosmetic
        self.customize_user_preferences_button.clicked.connect(self._open_user_preferences_dialog)

    @property
    def current_player_index(self) -> int:
        return self.player_index_combo.currentData()

    # Operations
    def _copy_permalink(self):
        QApplication.clipboard().setText(self.layout_description.permalink.as_base64_str)

    def _export_log(self):
        all_games: Set[RandovaniaGame] = {preset.game for preset in self.layout_description.permalink.presets.values()}
        if len(all_games) > 1:
            game_name = "Crossgame Multiworld"
        else:
            game_name = f"{list(all_games)[0].short_name} Randomizer"

        default_name = "{} - {}.{}".format(game_name,
                                           self.layout_description.shareable_word_hash,
                                           self.layout_description.file_extension())
        json_path = prompt_user_for_output_game_log(self, default_name=default_name)
        if json_path is not None:
            self.layout_description.save_to_file(json_path)

    def _export_preset(self):
        preset = self.layout_description.permalink.get_preset(self.current_player_index)
        output_path = common_qt_lib.prompt_user_for_preset_file(self, new_file=True)
        if output_path is not None:
            VersionedPreset.with_preset(preset).save_to_file(output_path)

    def _view_trick_usages(self):
        from randovania.gui.dialog.trick_usage_popup import TrickUsagePopup
        preset = self.layout_description.permalink.get_preset(self.current_player_index)
        self._trick_usage_popup = TrickUsagePopup(self, self._window_manager, preset)
        self._trick_usage_popup.setWindowModality(QtCore.Qt.WindowModal)
        self._trick_usage_popup.open()

    async def _show_dialog_for_prime3_layout(self):
        from randovania.games.patchers import gollop_corruption_patcher

        cosmetic = self._options.options_for_game(RandovaniaGame.METROID_PRIME_CORRUPTION).cosmetic_patches
        preset = self.layout_description.permalink.get_preset(self.current_player_index)
        patches = self.layout_description.all_patches[self.current_player_index]
        game = default_database.game_description_for(RandovaniaGame.METROID_PRIME_CORRUPTION)

        pickup_names = []
        for index in range(game.world_list.num_pickup_nodes):
            p_index = PickupIndex(index)
            if p_index in patches.pickup_assignment:
                name = patches.pickup_assignment[p_index].pickup.name
            else:
                name = "Missile Expansion"
            pickup_names.append(name)

        layout_string = gollop_corruption_patcher.layout_string_for_items(pickup_names)
        starting_location = patches.starting_location

        suit_type = game.resource_database.get_item_by_name("Suit Type")
        starting_items = copy.copy(patches.starting_items)
        starting_items[suit_type] = starting_items.get(suit_type, 0) + cosmetic.player_suit.value
        if preset.configuration.start_with_corrupted_hypermode:
            hypermode_original = 0
        else:
            hypermode_original = 1

        commands = "\n".join([
            f'set seed="{layout_string}"',
            f'set "starting_items={gollop_corruption_patcher.starting_items_for(starting_items, hypermode_original)}"',
            f'set "starting_location={gollop_corruption_patcher.starting_location_for(starting_location)}"',
            f'set "random_door_colors={str(cosmetic.random_door_colors).lower()}"',
            f'set "random_welding_colors={str(cosmetic.random_welding_colors).lower()}"',
        ])
        dialog_text = (
            "There is no integrated patcher for Metroid Prime 3: Corruption games.\n"
            "Download the randomizer for it from #corruption-general in the Metroid Prime Randomizer Discord, "
            "and use the following commands as a seed.\n\n"
            "\n{}").format(commands)

        message_box = ScrollLabelDialog(dialog_text, "Commands for patcher", self)
        message_box.resize(750, 200)
        message_box.label.setTextInteractionFlags(Qt.TextSelectableByMouse)
        QApplication.clipboard().setText(commands)
        await async_dialog.execute_dialog(message_box)

    @property
    def current_player_game(self) -> RandovaniaGame:
        return self.layout_description.permalink.get_preset(self.current_player_index).configuration.game

    @asyncSlot()
    async def _export_iso(self):
        layout = self.layout_description
        has_spoiler = layout.permalink.spoiler
        options = self._options

        if not options.is_alert_displayed(InfoAlert.FAQ):
            await async_dialog.message_box(self, QtWidgets.QMessageBox.Icon.Information, "FAQ",
                                           "Have you read the Randovania FAQ?\n"
                                           "It can be found in the main Randovania window → Help → FAQ")
            options.mark_alert_as_displayed(InfoAlert.FAQ)

        game = self.current_player_game
        if game == RandovaniaGame.METROID_PRIME_CORRUPTION:
            return await self._show_dialog_for_prime3_layout()

        patcher = self._window_manager.patcher_provider.patcher_for_game(game)

        dialog = GameInputDialog(options, patcher, layout.shareable_word_hash, has_spoiler, game)
        result = await async_dialog.execute_dialog(dialog)
        if result != QDialog.Accepted:
            return

        dialog.save_options()
        players_config = PlayersConfiguration(
            player_index=self.current_player_index,
            player_names=self._player_names,
        )

        patch_data = patcher.create_patch_data(layout, players_config, options.options_for_game(game).cosmetic_patches)
        self._can_stop_background_process = patcher.export_can_be_aborted
        await game_exporter.export_game(
            patcher=patcher,
            input_dialog=dialog,
            patch_data=patch_data,
            internal_copies_path=options.internal_copies_path,
            layout_for_spoiler=layout,
            background=self,
            progress_update_signal=self.progress_update_signal,
        )
        self._can_stop_background_process = True

    def _open_map_tracker(self):
        current_preset = self.layout_description.permalink.presets[self.current_player_index]
        self._window_manager.open_map_tracker(current_preset)

    # Layout Visualization
    def _create_pickup_spoilers(self, game_description: GameDescription):
        if game_description == self._pickup_spoiler_current_game:
            return

        for groups in self._pickup_spoiler_world_to_group.values():
            groups.deleteLater()

        self._pickup_spoiler_current_game = game_description
        self.pickup_spoiler_show_all_button.currently_show_all = True
        self.pickup_spoiler_buttons.clear()

        self._pickup_spoiler_world_to_group = {}
        nodes_in_world = collections.defaultdict(list)

        for world, area, node in game_description.world_list.all_worlds_areas_nodes:
            if isinstance(node, PickupNode):
                world_name = world.correct_name(area.in_dark_aether)
                nodes_in_world[world_name].append((f"{area.name} - {node.name}", node.pickup_index))
                continue

        for world_name in sorted(nodes_in_world.keys()):
            group_box = QGroupBox(self.pickup_spoiler_scroll_contents)
            group_box.setTitle(world_name)
            vertical_layout = QVBoxLayout(group_box)
            vertical_layout.setContentsMargins(8, 4, 8, 4)
            vertical_layout.setSpacing(2)
            group_box.vertical_layout = vertical_layout

            vertical_layout.horizontal_layouts = []
            self._pickup_spoiler_world_to_group[world_name] = group_box
            self.pickup_spoiler_scroll_content_layout.addWidget(group_box)

            for area_name, pickup_index in sorted(nodes_in_world[world_name], key=lambda it: it[0]):
                horizontal_layout = QHBoxLayout()
                horizontal_layout.setSpacing(2)

                label = QLabel(group_box)
                label.setText(area_name)
                horizontal_layout.addWidget(label)
                horizontal_layout.label = label

                push_button = QPushButton(group_box)
                push_button.setFlat(True)
                push_button.setText("Hidden")
                push_button.item_is_hidden = True
                push_button.pickup_index = pickup_index
                push_button.clicked.connect(partial(self._toggle_pickup_spoiler, push_button))
                push_button.item_name = "Nothing was Set, ohno"
                push_button.row = horizontal_layout
                horizontal_layout.addWidget(push_button)
                horizontal_layout.button = push_button
                self.pickup_spoiler_buttons.append(push_button)

                group_box.vertical_layout.addLayout(horizontal_layout)
                group_box.vertical_layout.horizontal_layouts.append(horizontal_layout)

    def create_history_item(self, node):
        button = QRadioButton(self.layout_history_content)
        button.toggled.connect(self.on_select_node)
        button.setText(node)
        self.layout_history_content_layout.addWidget(button)
        self._history_items.append(button)
        return button

    def update_layout_description(self, description: LayoutDescription):
        self.layout_description = description
        self.layout_info_tab.show()

        self.setWindowTitle(f"Game Details: {description.shareable_word_hash}")
        self.export_log_button.setEnabled(description.permalink.spoiler)

        self._player_names = {
            i: f"Player {i + 1}"
            for i in range(description.permalink.player_count)
        }

        self.export_iso_button.setEnabled(description.permalink.player_count == 1)
        if description.permalink.player_count > 1:
            self.export_iso_button.setToolTip("Multiworld games can only be exported from a game session")
        else:
            self.export_iso_button.setToolTip("")

        self.customize_user_preferences_button.setVisible(description.permalink.player_count == 1)

        self.player_index_combo.clear()
        for i in range(description.permalink.player_count):
            self.player_index_combo.addItem(self._player_names[i], i)
        self.player_index_combo.setCurrentIndex(0)
        self.player_index_combo.setVisible(description.permalink.player_count > 1)

        self._update_current_player()

    def _update_current_player(self):
        description = self.layout_description
        current_player = self.current_player_index
        preset = description.permalink.get_preset(current_player)

        self.permalink_edit.setText(description.permalink.as_base64_str)
        title_text = """
        <p>
            Seed Hash: {description.shareable_word_hash} ({description.shareable_hash})<br/>
            Preset Name: {preset.name}
        </p>
        """.format(description=description, preset=preset)
        self.layout_title_label.setText(title_text)

        categories = list(preset_describer.describe(preset))
        self.layout_description_left_label.setText(preset_describer.merge_categories(categories[::2]))
        self.layout_description_right_label.setText(preset_describer.merge_categories(categories[1::2]))

        # Game Spoiler
        has_spoiler = description.permalink.spoiler
        self.pickup_tab.setEnabled(has_spoiler)
        patches = description.all_patches[current_player]

        if has_spoiler:
            pickup_names = {
                pickup.pickup.name
                for pickup in patches.pickup_assignment.values()
            }
            game_description = default_database.game_description_for(preset.game)
            self._create_pickup_spoilers(game_description)
            starting_area = game_description.world_list.area_by_area_location(patches.starting_location)

            extra_items = item_names.additional_starting_items(preset.configuration,
                                                               game_description.resource_database,
                                                               patches.starting_items)

            self.spoiler_starting_location_label.setText("Starting Location: {}".format(
                game_description.world_list.area_name(starting_area)
            ))
            self.spoiler_starting_items_label.setText("Random Starting Items: {}".format(
                ", ".join(extra_items)
                if extra_items else "None"
            ))
            self._update_show_all_button_state()

        else:
            pickup_names = {}
            self.layout_info_tab.removeTab(self.layout_info_tab.indexOf(self.pickup_tab))
            self.spoiler_starting_location_label.setText("Starting Location")
            self.spoiler_starting_items_label.setText("Random Starting Items")

        self.pickup_spoiler_pickup_combobox.clear()
        self.pickup_spoiler_pickup_combobox.addItem("None")
        for pickup_name in sorted(pickup_names):
            self.pickup_spoiler_pickup_combobox.addItem(pickup_name)

        for pickup_button in self.pickup_spoiler_buttons:
            pickup_target = patches.pickup_assignment.get(pickup_button.pickup_index)

            pickup_button.target_player = None
            if pickup_target is not None:
                pickup_button.item_name = pickup_target.pickup.name if has_spoiler else "????"
                if has_spoiler and description.permalink.player_count > 1:
                    pickup_button.target_player = pickup_target.player
                    pickup_button.player_names = self._player_names
            else:
                pickup_button.item_name = "Nothing"

            if not pickup_button.item_is_hidden:
                pickup_button.setText(pickup_button.item_name)

    def _toggle_pickup_spoiler(self, button):
        if button.item_is_hidden:
            _show_pickup_spoiler(button)
        else:
            _hide_pickup_spoiler(button)
        self._update_show_all_button_state()

    def _toggle_show_all_pickup_spoiler(self):
        if self.pickup_spoiler_show_all_button.currently_show_all:
            action = _show_pickup_spoiler
        else:
            action = _hide_pickup_spoiler

        for button in self.pickup_spoiler_buttons:
            action(button)

        self._update_show_all_button_state()

    def _update_show_all_button_state(self):
        self.pickup_spoiler_show_all_button.currently_show_all = all(
            button.item_is_hidden for button in self.pickup_spoiler_buttons
        )
        if self.pickup_spoiler_show_all_button.currently_show_all:
            self.pickup_spoiler_show_all_button.setText(QtCore.QCoreApplication.translate("HistoryWindow", "Show All"))
        else:
            self.pickup_spoiler_show_all_button.setText(QtCore.QCoreApplication.translate("HistoryWindow", "Hide All"))

    def _on_change_pickup_filter(self, text):
        for button in self.pickup_spoiler_buttons:
            visible = text == "None" or text == button.item_name
            button.setVisible(visible)
            button.row.label.setVisible(visible)

    def _open_user_preferences_dialog(self):
        game = self.current_player_game
        per_game_options = self._options.options_for_game(game)

        dialog = game_specific_gui.create_dialog_for_cosmetic_patches(self, per_game_options.cosmetic_patches)
        result = dialog.exec_()
        if result == QtWidgets.QDialog.Accepted:
            with self._options as options:
                options.set_options_for_game(game, dataclasses.replace(per_game_options,
                                                                       cosmetic_patches=dialog.cosmetic_patches))

    def enable_buttons_with_background_tasks(self, value: bool):
        self.stop_background_process_button.setEnabled(not value and self._can_stop_background_process)
        self.export_iso_button.setEnabled(value)
        simplified_patcher.export_busy = not value

    def update_progress(self, message: str, percentage: int):
        self.progress_label.setText(message)
        if "Aborted" in message:
            percentage = 0

        if percentage is None:
            percentage = self._last_percentage
        else:
            self._last_percentage = percentage
        if percentage >= 0:
            self.progress_bar.setRange(0, 100)
            self.progress_bar.setValue(percentage)
        else:
            self.progress_bar.setRange(0, 0)

    def ignore_close_event(self, event: QtGui.QCloseEvent) -> bool:
        return not self._can_stop_background_process



import collections
import re
from typing import List, DefaultDict, Dict, FrozenSet, Tuple, Iterator, Set

from randovania.game_description.assignment import PickupTarget
from randovania.game_description.game_description import GameDescription
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.world.node import ResourceNode
from randovania.game_description.world.world import World
from randovania.game_description.world.world_list import WorldList
from randovania.generator import reach_lib
from randovania.generator.filler.action import Action
from randovania.generator.filler.filler_configuration import FillerConfiguration
from randovania.generator.filler.filler_library import UncollectedState
from randovania.generator.filler.filler_logging import print_new_resources, print_retcon_loop_start
from randovania.generator.filler.pickup_list import (get_pickups_that_solves_unreachable,
                                                     get_pickups_with_interesting_resources,
                                                     interesting_resources_for_reach, PickupCombinations)
from randovania.layout.base.available_locations import RandomizationMode
from randovania.layout.base.logical_resource_action import LayoutLogicalResourceAction
from randovania.resolver import debug
from randovania.resolver.state import State


class PlayerState:
    index: int
    game: GameDescription
    pickups_left: List[PickupEntry]
    configuration: FillerConfiguration
    pickup_index_seen_count: DefaultDict[PickupIndex, int]
    scan_asset_seen_count: DefaultDict[LogbookAsset, int]
    scan_asset_initial_pickups: Dict[LogbookAsset, FrozenSet[PickupIndex]]
    _unfiltered_potential_actions: Tuple[PickupCombinations, Tuple[ResourceNode, ...]]
    num_random_starting_items_placed: int
    num_assigned_pickups: int

    def __init__(self,
                 index: int,
                 game: GameDescription,
                 initial_state: State,
                 pickups_left: List[PickupEntry],
                 configuration: FillerConfiguration,
                 ):
        self.index = index
        self.game = game

        self.reach = reach_lib.advance_reach_with_possible_unsafe_resources(
            reach_lib.reach_with_all_safe_resources(game, initial_state))
        self.pickups_left = pickups_left
        self.configuration = configuration

        self.pickup_index_seen_count = collections.defaultdict(int)
        self.scan_asset_seen_count = collections.defaultdict(int)
        self.event_seen_count = collections.defaultdict(int)
        self.scan_asset_initial_pickups = {}
        self.num_random_starting_items_placed = 0
        self.num_assigned_pickups = 0
        self.num_actions = 0
        self.indices_groups, self.all_indices = build_available_indices(game.world_list, configuration)

    def __repr__(self):
        return f"Player {self.index + 1}"

    def update_for_new_state(self):
        debug.debug_print(f"\n>>> Updating state of {self}")
        self._advance_pickup_index_seen_count()
        self._advance_scan_asset_seen_count()
        self._advance_event_seen_count()
        self._calculate_potential_actions()

    def _advance_pickup_index_seen_count(self):
        for pickup_index in self.reach.state.collected_pickup_indices:
            self.pickup_index_seen_count[pickup_index] += 1

        print_new_resources(self.game, self.reach, self.pickup_index_seen_count, "Pickup Index")

    def _advance_scan_asset_seen_count(self):
        for scan_asset in self.reach.state.collected_scan_assets:
            self.scan_asset_seen_count[scan_asset] += 1
            if self.scan_asset_seen_count[scan_asset] == 1:
                self.scan_asset_initial_pickups[scan_asset] = frozenset(self.reach.state.collected_pickup_indices)

        print_new_resources(self.game, self.reach, self.scan_asset_seen_count, "Scan Asset")

    def _advance_event_seen_count(self):
        for resource, quantity in self.reach.state.resources.items():
            if resource.resource_type == ResourceType.EVENT and quantity > 0:
                self.event_seen_count[resource] += 1

        print_new_resources(self.game, self.reach, self.event_seen_count, "Events")

    def _calculate_potential_actions(self):
        uncollected_resource_nodes = reach_lib.get_collectable_resource_nodes_of_reach(self.reach)
        if self.configuration.multi_pickup_placement:
            get_pickups = get_pickups_that_solves_unreachable
        else:
            get_pickups = get_pickups_with_interesting_resources

        usable_pickups = [pickup for pickup in self.pickups_left
                          if self.num_actions >= pickup.required_progression]
        pickups = get_pickups(usable_pickups, self.reach, uncollected_resource_nodes)
        print_retcon_loop_start(self.game, usable_pickups, self.reach, self.index)

        self._unfiltered_potential_actions = pickups, tuple(uncollected_resource_nodes)

    def potential_actions(self, num_available_indices: int) -> List[Action]:
        num_available_indices += (self.configuration.maximum_random_starting_items
                                  - self.num_random_starting_items_placed)

        pickups, uncollected_resource_nodes = self._unfiltered_potential_actions
        result: List[Action] = [pickup_tuple for pickup_tuple in pickups
                                if len(pickup_tuple) <= num_available_indices]

        logical_resource_action = self.configuration.logical_resource_action
        if (logical_resource_action == LayoutLogicalResourceAction.RANDOMLY
                or (logical_resource_action == LayoutLogicalResourceAction.LAST_RESORT and not result)):
            result.extend(uncollected_resource_nodes)

        return result

    def victory_condition_satisfied(self):
        return self.game.victory_condition.satisfied(self.reach.state.resources,
                                                     self.reach.state.energy,
                                                     self.reach.state.resource_database)

    def assign_pickup(self, pickup_index: PickupIndex, target: PickupTarget):
        self.num_assigned_pickups += 1
        self.reach.state.patches = self.reach.state.patches.assign_new_pickups([
            (pickup_index, target),
        ])

    def current_state_report(self) -> str:
        state = UncollectedState.from_reach(self.reach)
        pickups_by_name_and_quantity = collections.defaultdict(int)

        _KEY_MATCH = re.compile(r"Key (\d+)")
        for pickup in self.pickups_left:
            pickups_by_name_and_quantity[_KEY_MATCH.sub("Key", pickup.name)] += 1

        to_progress = {_KEY_MATCH.sub("Key", resource.long_name)
                       for resource in interesting_resources_for_reach(self.reach)
                       if resource.resource_type == ResourceType.ITEM}

        return ("At {0} after {1} actions and {2} pickups, with {3} collected locations.\n\n"
                "Pickups still available: {4}\n\nResources to progress: {5}").format(
            self.game.world_list.node_name(self.reach.state.node, with_world=True, distinguish_dark_aether=True),
            self.num_actions,
            self.num_assigned_pickups,
            len(state.indices),
            ", ".join(name if quantity == 1 else f"{name} x{quantity}"
                      for name, quantity in sorted(pickups_by_name_and_quantity.items())),
            ", ".join(sorted(to_progress)),
        )


def world_indices_for_mode(world: World, randomization_mode: RandomizationMode) -> Iterator[PickupIndex]:
    if randomization_mode is RandomizationMode.FULL:
        yield from world.pickup_indices
    elif randomization_mode is RandomizationMode.MAJOR_MINOR_SPLIT:
        yield from world.major_pickup_indices
    else:
        raise RuntimeError("Unknown randomization_mode: {}".format(randomization_mode))


def build_available_indices(world_list: WorldList, configuration: FillerConfiguration,
                            ) -> Tuple[List[Set[PickupIndex]], Set[PickupIndex]]:
    """
    Groups indices into separated groups, so each group can be weighted separately.
    """
    indices_groups = [
        set(world_indices_for_mode(world, configuration.randomization_mode)) - configuration.indices_to_exclude
        for world in world_list.worlds
    ]
    all_indices = set().union(*indices_groups)

    return indices_groups, all_indices



from randovania.game_description.item.item_category import ItemCategory
from typing import Iterator, Tuple

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackFloat, BitPackDecoder
from randovania.game_description import default_database
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry, ResourceConversion, ResourceLock, \
    PickupModel
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceQuantity
from randovania.games.game import RandovaniaGame

_PROBABILITY_OFFSET_META = {
    "min": -3,
    "max": 3,
    "precision": 2.0,
    "if_different": 0.0,
}
_PROBABILITY_MULTIPLIER_META = {
    "min": 0,
    "max": 8,
    "precision": 2.0,
    "if_different": 1.0,
}


class DatabaseBitPackHelper:
    def __init__(self, database: ResourceDatabase):
        self.database = database

    def _decode_item(self, decoder: BitPackDecoder) -> ItemResourceInfo:
        return decoder.decode_element(self.database.item)

    # Resource Quantity
    def encode_resource_quantity(self, item: ResourceQuantity):
        yield from bitpacking.pack_array_element(item[0], self.database.item)
        assert item[1] <= item[0].max_capacity
        yield item[1], item[0].max_capacity + 1

    def decode_resource_quantity(self, decoder: BitPackDecoder) -> ResourceQuantity:
        resource = self._decode_item(decoder)
        quantity = decoder.decode_single(resource.max_capacity + 1)
        return resource, quantity

    # Resource Conversion
    def encode_resource_conversion(self, item: ResourceConversion):
        yield from bitpacking.pack_array_element(item.source, self.database.item)
        yield from bitpacking.pack_array_element(item.target, self.database.item)

    def decode_resource_conversion(self, decoder: BitPackDecoder) -> ResourceConversion:
        return ResourceConversion(
            source=self._decode_item(decoder),
            target=self._decode_item(decoder),
        )

    # Resource Lock
    def encode_resource_lock(self, lock: ResourceLock):
        yield from bitpacking.pack_array_element(lock.locked_by, self.database.item)
        yield from bitpacking.pack_array_element(lock.item_to_lock, self.database.item)
        yield from bitpacking.pack_array_element(lock.temporary_item, self.database.item)

    def decode_resource_lock(self, decoder: BitPackDecoder) -> ResourceLock:
        return ResourceLock(
            locked_by=self._decode_item(decoder),
            item_to_lock=self._decode_item(decoder),
            temporary_item=self._decode_item(decoder),
        )

# Item categories encoding & decoding
def _encode_item_category(category: ItemCategory):
    yield from bitpacking.encode_string(category.name)
    yield from bitpacking.encode_string(category.long_name)
    yield from bitpacking.encode_string(category.hint_details[0])
    yield from bitpacking.encode_string(category.hint_details[1])
    yield from bitpacking.encode_bool(category.is_major)
    yield from bitpacking.encode_bool(category.is_key)

def _decode_item_category(decoder: BitPackDecoder) -> ItemCategory:
    return ItemCategory(
        name=bitpacking.decode_string(decoder),
        long_name=bitpacking.decode_string(decoder),
        hint_details=(bitpacking.decode_string(decoder), bitpacking.decode_string(decoder)),
        is_major=bitpacking.decode_bool(decoder),
        is_key=bitpacking.decode_bool(decoder)
    )


class BitPackPickupEntry:
    value: PickupEntry
    database: ResourceDatabase
    
    def __init__(self, value: PickupEntry, database: ResourceDatabase):
        self.value = value
        self.database = database

    # Main Methods
    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        helper = DatabaseBitPackHelper(self.database)

        yield from bitpacking.encode_string(self.value.name)
        yield from self.value.model.game.bit_pack_encode({})
        yield from bitpacking.encode_string(self.value.model.name)
        yield from _encode_item_category(self.value.item_category)
        yield from _encode_item_category(self.value.broad_category)
        yield from bitpacking.encode_tuple(self.value.progression, helper.encode_resource_quantity)
        yield from bitpacking.encode_tuple(self.value.extra_resources, helper.encode_resource_quantity)
        yield from bitpacking.encode_bool(self.value.unlocks_resource)
        yield from bitpacking.encode_bool(self.value.resource_lock is not None)
        if self.value.resource_lock is not None:
            yield from helper.encode_resource_lock(self.value.resource_lock)
        yield from bitpacking.encode_bool(self.value.respects_lock)
        yield from BitPackFloat(self.value.probability_offset).bit_pack_encode(_PROBABILITY_OFFSET_META)
        yield from BitPackFloat(self.value.probability_multiplier).bit_pack_encode(_PROBABILITY_MULTIPLIER_META)
        yield from bitpacking.encode_big_int(self.value.required_progression)

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, database: ResourceDatabase) -> PickupEntry:
        helper = DatabaseBitPackHelper(database)

        name = bitpacking.decode_string(decoder)
        model = PickupModel(
            game=RandovaniaGame.bit_pack_unpack(decoder, {}),
            name=bitpacking.decode_string(decoder),
        )
        item_category = _decode_item_category(decoder)
        broad_category = _decode_item_category(decoder)
        progression = bitpacking.decode_tuple(decoder, helper.decode_resource_quantity)
        extra_resources = bitpacking.decode_tuple(decoder, helper.decode_resource_quantity)
        unlocks_resource = bitpacking.decode_bool(decoder)
        resource_lock = None
        if bitpacking.decode_bool(decoder):
            resource_lock = helper.decode_resource_lock(decoder)
        respects_lock = bitpacking.decode_bool(decoder)
        probability_offset = BitPackFloat.bit_pack_unpack(decoder, _PROBABILITY_OFFSET_META)
        probability_multiplier = BitPackFloat.bit_pack_unpack(decoder, _PROBABILITY_MULTIPLIER_META)
        required_progression = bitpacking.decode_big_int(decoder)

        return PickupEntry(
            name=name,
            model=model,
            item_category=item_category,
            broad_category=broad_category,
            progression=progression,
            extra_resources=extra_resources,
            unlocks_resource=unlocks_resource,
            resource_lock=resource_lock,
            respects_lock=respects_lock,
            probability_offset=probability_offset,
            probability_multiplier=probability_multiplier,
            required_progression=required_progression,
        )



from typing import List

import pytest

from randovania.game_description import default_database
from randovania.game_description.assignment import PickupTarget
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.games.game import RandovaniaGame
from randovania.games.prime import echoes_items
from randovania.games.prime.patcher_file_lib import sky_temple_key_hint, hint_lib
from randovania.games.prime.patcher_file_lib.hint_name_creator import LocationHintCreator
from randovania.generator.item_pool import pickup_creator
from randovania.interface_common.players_configuration import PlayersConfiguration


def _create_hint_text(hide_area: bool,
                      multiworld: bool,
                      key_number: int,
                      key_location: str,
                      ) -> List[str]:
    if hide_area:
        key_location = key_location.split(" - ")[0]

    determiner = ""
    if multiworld:
        player = "you" if key_number < 6 else "them"
        determiner = f"&push;&main-color=#d4cc33;{player}&pop;'s "

    message = (f"&push;&main-color=#FF6705B3;Sky Temple Key {key_number}&pop; is located in "
               f"{determiner}&push;&main-color=#FF3333;{key_location}&pop;.")
    return [message, "", message]


def make_starting_stk_hint(key_number: int) -> List[str]:
    message = f"&push;&main-color=#FF6705B3;Sky Temple Key {key_number}&pop; has no need to be located."
    return [message, "", message]


def make_useless_stk_hint(key_number: int) -> List[str]:
    message = f"&push;&main-color=#FF6705B3;Sky Temple Key {key_number}&pop; is lost somewhere in Aether."
    return [message, "", message]


@pytest.mark.parametrize("multiworld", [False, True])
@pytest.mark.parametrize("hide_area", [False, True])
def test_create_hints_all_placed(hide_area: bool, multiworld: bool, empty_patches):
    # Setup
    prime_game = default_database.game_description_for(RandovaniaGame.METROID_PRIME)
    echoes_game = default_database.game_description_for(RandovaniaGame.METROID_PRIME_ECHOES)

    players_config = PlayersConfiguration(0, {0: "you", 1: "them"} if multiworld else {0: "you"})
    area_namers = {
        0: hint_lib.AreaNamer(echoes_game.world_list),
        1: hint_lib.AreaNamer(prime_game.world_list),
    }

    patches = empty_patches.assign_new_pickups([
        (PickupIndex(17 + key),
         PickupTarget(pickup_creator.create_sky_temple_key(key, echoes_game.resource_database), 0))
        for key in range(5 if multiworld else 9)
    ])
    other_patches = empty_patches.assign_new_pickups([
        (PickupIndex(17 + key),
         PickupTarget(pickup_creator.create_sky_temple_key(key, echoes_game.resource_database), 0))
        for key in range(5, 9)
    ])
    assets = [0xD97685FE, 0x32413EFD, 0xDD8355C3, 0x3F5F4EBA, 0xD09D2584,
              0x3BAA9E87, 0xD468F5B9, 0x2563AE34, 0xCAA1C50A]

    locations = [
        "Sky Temple Grounds - Profane Path",
        "Sky Temple Grounds - Phazon Grounds",
        "Sky Temple Grounds - Ing Reliquary",
        "Great Temple - Transport A Access",
        "Great Temple - Temple Sanctuary",
        "Great Temple - Transport B Access",
        "Great Temple - Main Energy Controller",
        "Great Temple - Main Energy Controller",
        "Agon Wastes - Mining Plaza",
    ]
    other_locations = [
        "Chozo Ruins - Watery Hall Access",
        "Chozo Ruins - Watery Hall",
        "Chozo Ruins - Watery Hall",
        "Chozo Ruins - Dynamo",
    ]
    if multiworld:
        locations[-len(other_locations):] = other_locations

    expected = [
        {"asset_id": asset_id, "strings": _create_hint_text(hide_area, multiworld, i + 1, text)}
        for i, (asset_id, text) in enumerate(zip(assets, locations))
    ]

    # Run
    result = sky_temple_key_hint.create_hints(
        {0: patches, 1: other_patches} if multiworld else {0: patches},
        players_config,
        echoes_game.resource_database,
        area_namers,
        hide_area)

    # Assert
    assert result == expected


@pytest.mark.parametrize("multiworld", [False, True])
@pytest.mark.parametrize("hide_area", [False, True])
def test_create_hints_all_starting(hide_area: bool, multiworld: bool,
                                   empty_patches, echoes_game_description):
    # Setup
    players_config = PlayersConfiguration(0, {0: "you", 1: "them"} if multiworld else {0: "you"})
    area_namer = {0: None, 1: None} if multiworld else {0: None}
    patches = empty_patches.assign_extra_initial_items({
        echoes_game_description.resource_database.get_item(echoes_items.SKY_TEMPLE_KEY_ITEMS[key]): 1
        for key in range(9)
    })

    expected = [
        {"asset_id": 0xD97685FE,
         "strings": make_starting_stk_hint(1)},
        {"asset_id": 0x32413EFD,
         "strings": make_starting_stk_hint(2)},
        {"asset_id": 0xDD8355C3,
         "strings": make_starting_stk_hint(3)},
        {"asset_id": 0x3F5F4EBA,
         "strings": make_starting_stk_hint(4)},
        {"asset_id": 0xD09D2584,
         "strings": make_starting_stk_hint(5)},
        {"asset_id": 0x3BAA9E87,
         "strings": make_starting_stk_hint(6)},
        {"asset_id": 0xD468F5B9,
         "strings": make_starting_stk_hint(7)},
        {"asset_id": 0x2563AE34,
         "strings": make_starting_stk_hint(8)},
        {"asset_id": 0xCAA1C50A,
         "strings": make_starting_stk_hint(9)},
    ]

    # Run
    result = sky_temple_key_hint.create_hints({0: patches}, players_config, echoes_game_description.resource_database,
                                              area_namer, hide_area)

    # Assert
    assert result == expected


def test_hide_hints():
    # Setup
    expected = [
        {"asset_id": 0xD97685FE,
         "strings": make_useless_stk_hint(1)},
        {"asset_id": 0x32413EFD,
         "strings": make_useless_stk_hint(2)},
        {"asset_id": 0xDD8355C3,
         "strings": make_useless_stk_hint(3)},
        {"asset_id": 0x3F5F4EBA,
         "strings": make_useless_stk_hint(4)},
        {"asset_id": 0xD09D2584,
         "strings": make_useless_stk_hint(5)},
        {"asset_id": 0x3BAA9E87,
         "strings": make_useless_stk_hint(6)},
        {"asset_id": 0xD468F5B9,
         "strings": make_useless_stk_hint(7)},
        {"asset_id": 0x2563AE34,
         "strings": make_useless_stk_hint(8)},
        {"asset_id": 0xCAA1C50A,
         "strings": make_useless_stk_hint(9)},
    ]

    # Run
    result = sky_temple_key_hint.hide_hints()

    # Assert
    assert result == expected



import dataclasses

from randovania.bitpacking.json_dataclass import JsonDataclass


@dataclasses.dataclass(frozen=True)
class BaseCosmeticPatches(JsonDataclass):
    pass

    @classmethod
    def default(cls) -> "BaseCosmeticPatches":
        return cls()

    @classmethod
    def game(cls):
        raise NotImplementedError()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_item_pool.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from randovania.gui.lib.scroll_protected import ScrollProtectedSpinBox


class Ui_PresetItemPool(object):
    def setupUi(self, PresetItemPool):
        if not PresetItemPool.objectName():
            PresetItemPool.setObjectName(u"PresetItemPool")
        PresetItemPool.resize(566, 227)
        PresetItemPool.setMaximumSize(QSize(16777215, 16777215))
        self.centralWidget = QWidget(PresetItemPool)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.verticalLayout = QVBoxLayout(self.centralWidget)
        self.verticalLayout.setSpacing(6)
        self.verticalLayout.setContentsMargins(11, 11, 11, 11)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(5, 5, 5, 5)
        self.item_pool_count_label = QLabel(self.centralWidget)
        self.item_pool_count_label.setObjectName(u"item_pool_count_label")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.item_pool_count_label.sizePolicy().hasHeightForWidth())
        self.item_pool_count_label.setSizePolicy(sizePolicy)
        font = QFont()
        font.setPointSize(12)
        self.item_pool_count_label.setFont(font)
        self.item_pool_count_label.setAlignment(Qt.AlignCenter)

        self.verticalLayout.addWidget(self.item_pool_count_label)

        self.verticalSpacer = QSpacerItem(20, 3, QSizePolicy.Minimum, QSizePolicy.Minimum)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.scroll_area = QScrollArea(self.centralWidget)
        self.scroll_area.setObjectName(u"scroll_area")
        self.scroll_area.setFrameShape(QFrame.NoFrame)
        self.scroll_area.setFrameShadow(QFrame.Plain)
        self.scroll_area.setLineWidth(0)
        self.scroll_area.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scroll_area.setWidgetResizable(True)
        self.scroll_area.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.scroll_area_contents = QWidget()
        self.scroll_area_contents.setObjectName(u"scroll_area_contents")
        self.scroll_area_contents.setGeometry(QRect(0, 0, 556, 183))
        self.item_pool_layout = QVBoxLayout(self.scroll_area_contents)
        self.item_pool_layout.setSpacing(6)
        self.item_pool_layout.setContentsMargins(11, 11, 11, 11)
        self.item_pool_layout.setObjectName(u"item_pool_layout")
        self.item_pool_layout.setContentsMargins(0, 6, 0, 0)
        self.multi_pickup_placement_check = QCheckBox(self.scroll_area_contents)
        self.multi_pickup_placement_check.setObjectName(u"multi_pickup_placement_check")

        self.item_pool_layout.addWidget(self.multi_pickup_placement_check)

        self.random_starting_box = QGroupBox(self.scroll_area_contents)
        self.random_starting_box.setObjectName(u"random_starting_box")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.random_starting_box.sizePolicy().hasHeightForWidth())
        self.random_starting_box.setSizePolicy(sizePolicy1)
        self.gridLayout_2 = QGridLayout(self.random_starting_box)
        self.gridLayout_2.setSpacing(6)
        self.gridLayout_2.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.minimum_starting_spinbox = ScrollProtectedSpinBox(self.random_starting_box)
        self.minimum_starting_spinbox.setObjectName(u"minimum_starting_spinbox")
        self.minimum_starting_spinbox.setMaximum(30)

        self.gridLayout_2.addWidget(self.minimum_starting_spinbox, 1, 1, 1, 1)

        self.minimum_starting_label = QLabel(self.random_starting_box)
        self.minimum_starting_label.setObjectName(u"minimum_starting_label")
        self.minimum_starting_label.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignVCenter)

        self.gridLayout_2.addWidget(self.minimum_starting_label, 1, 0, 1, 1)

        self.random_starting_label = QLabel(self.random_starting_box)
        self.random_starting_label.setObjectName(u"random_starting_label")
        self.random_starting_label.setFrameShape(QFrame.NoFrame)
        self.random_starting_label.setFrameShadow(QFrame.Plain)
        self.random_starting_label.setWordWrap(True)

        self.gridLayout_2.addWidget(self.random_starting_label, 0, 0, 1, 2)

        self.maximum_starting_spinbox = ScrollProtectedSpinBox(self.random_starting_box)
        self.maximum_starting_spinbox.setObjectName(u"maximum_starting_spinbox")
        self.maximum_starting_spinbox.setMaximum(30)

        self.gridLayout_2.addWidget(self.maximum_starting_spinbox, 2, 1, 1, 1)

        self.maximum_starting_label = QLabel(self.random_starting_box)
        self.maximum_starting_label.setObjectName(u"maximum_starting_label")

        self.gridLayout_2.addWidget(self.maximum_starting_label, 2, 0, 1, 1)


        self.item_pool_layout.addWidget(self.random_starting_box)

        self.scroll_area.setWidget(self.scroll_area_contents)

        self.verticalLayout.addWidget(self.scroll_area)

        PresetItemPool.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetItemPool)

        QMetaObject.connectSlotsByName(PresetItemPool)
    # setupUi

    def retranslateUi(self, PresetItemPool):
        PresetItemPool.setWindowTitle(QCoreApplication.translate("PresetItemPool", u"Item Pool", None))
#if QT_CONFIG(tooltip)
        self.item_pool_count_label.setToolTip(QCoreApplication.translate("PresetItemPool", u"If there are fewer than 119 items, the rest of the item locations will contain Energy Transfer Modules.", None))
#endif // QT_CONFIG(tooltip)
        self.item_pool_count_label.setText(QCoreApplication.translate("PresetItemPool", u"Items in pool: #/119", None))
#if QT_CONFIG(tooltip)
        self.multi_pickup_placement_check.setToolTip(QCoreApplication.translate("PresetItemPool", u"<html><head/><body><p>Use the new version of the pickup placement logic, which is able to place multiple pickups at once to solve a requirement, like both Seeker Launcher and Dark Visor. Not tested yet.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.multi_pickup_placement_check.setText(QCoreApplication.translate("PresetItemPool", u"Use multi-pickup placement [Experimental]", None))
        self.random_starting_box.setTitle(QCoreApplication.translate("PresetItemPool", u"Random Starting Items", None))
        self.minimum_starting_label.setText(QCoreApplication.translate("PresetItemPool", u"Start with at least this many items:", None))
        self.random_starting_label.setText(QCoreApplication.translate("PresetItemPool", u"<html><head/><body><p>Randovania will add additional starting items if necessary to make the seed possible.<br/>The first value controls how many items are always added.<br/>The second value controls how many items the seed can have before it fails to generate.</p></body></html>", None))
        self.maximum_starting_label.setText(QCoreApplication.translate("PresetItemPool", u"Start with at most this many items:", None))
    # retranslateUi




import copy
import dataclasses
from typing import Tuple

from randovania.game_description import default_database
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.requirements import ResourceRequirement
from randovania.game_description.resources.damage_resource_info import DamageReduction
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import CurrentResources, \
    add_resource_gain_to_current_resources
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.world.node import PlayerShipNode
from randovania.games.game import RandovaniaGame
from randovania.layout.base.major_items_configuration import MajorItemsConfiguration
from randovania.layout.base.trick_level import LayoutTrickLevel
from randovania.layout.base.trick_level_configuration import TrickLevelConfiguration
from randovania.layout.game_to_class import AnyGameConfiguration
from randovania.resolver.state import State, StateGameData


def trick_resources_for_configuration(configuration: TrickLevelConfiguration,
                                      resource_database: ResourceDatabase,
                                      ) -> CurrentResources:
    """
    :param configuration:
    :param resource_database:
    :return:
    """

    static_resources = {}

    for trick in resource_database.trick:
        if configuration.minimal_logic:
            level = LayoutTrickLevel.maximum()
        else:
            level = configuration.level_for_trick(trick)
        static_resources[trick] = level.as_number

    return static_resources


def _add_minimal_logic_initial_resources(resources: CurrentResources,
                                         game: GameDescription,
                                         major_items: MajorItemsConfiguration,
                                         ) -> None:
    resource_database = game.resource_database

    if game.minimal_logic is None:
        raise ValueError(f"Minimal logic enabled, but {game.game} doesn't have support for it.")

    item_db = default_database.item_database_for_game(game.game)

    items_to_skip = set()
    for it in game.minimal_logic.items_to_exclude:
        if it.reason is None or major_items.items_state[item_db.major_items[it.reason]].num_shuffled_pickups != 0:
            items_to_skip.add(it.index)

    custom_item_count = game.minimal_logic.custom_item_amount
    events_to_skip = {it.index for it in game.minimal_logic.events_to_exclude}

    for event in resource_database.event:
        if event.index not in events_to_skip:
            resources[event] = 1

    for item in resource_database.item:
        if item.index not in items_to_skip:
            resources[item] = custom_item_count.get(item.index, 1)


def calculate_starting_state(game: GameDescription, patches: GamePatches, energy_per_tank: int) -> "State":
    starting_node = game.world_list.resolve_teleporter_connection(patches.starting_location)
    initial_resources = copy.copy(patches.starting_items)

    if isinstance(starting_node, PlayerShipNode):
        add_resource_gain_to_current_resources(
            starting_node.resource_gain_on_collect(patches, initial_resources, game.world_list.all_nodes,
                                                   game.resource_database),
            initial_resources,
        )

    initial_game_state = game.initial_states.get("Default")
    if initial_game_state is not None:
        add_resource_gain_to_current_resources(initial_game_state, initial_resources)

    starting_state = State(
        initial_resources,
        (),
        99 + (100 * initial_resources.get(game.resource_database.energy_tank, 0)),
        starting_node,
        patches,
        None,
        StateGameData(
            game.resource_database,
            game.world_list,
            energy_per_tank,
        )
    )

    # Being present with value 0 is troublesome since this dict is used for a simplify_requirements later on
    keys_to_remove = [resource for resource, quantity in initial_resources.items() if quantity == 0]
    for resource in keys_to_remove:
        del initial_resources[resource]

    return starting_state


def version_resources_for_game(resource_database: ResourceDatabase) -> CurrentResources:
    # All version differences are patched out from the game
    return {
        resource: 1 if resource.long_name == "NTSC" else 0
        for resource in resource_database.version
    }


def misc_resources_for_configuration(configuration: AnyGameConfiguration,
                                     resource_database: ResourceDatabase) -> CurrentResources:
    enabled_resources = set()
    if configuration.game == RandovaniaGame.METROID_PRIME:
        logical_patches = {
            "allow_underwater_movement_without_gravity": 0,
            "main_plaza_door": 1,
            "backwards_frigate": 2,
            "backwards_labs": 3,
            "backwards_upper_mines": 4,
            "backwards_lower_mines": 5,
            "phazon_elite_without_dynamo": 6,
            "small_samus": 7,
        }
        for name, index in logical_patches.items():
            if getattr(configuration, name):
                enabled_resources.add(index)

    elif configuration.game == RandovaniaGame.METROID_PRIME_ECHOES:
        enabled_resources = set()
        allow_vanilla = {
            "allow_jumping_on_dark_water": 1,
            "allow_vanilla_dark_beam": 19,
            "allow_vanilla_light_beam": 20,
            "allow_vanilla_seeker_launcher": 21,
            "allow_vanilla_echo_visor": 22,
            "allow_vanilla_dark_visor": 27,
            "allow_vanilla_screw_attack": 23,
            "allow_vanilla_gravity_boost": 24,
            "allow_vanilla_boost_ball": 25,
            "allow_vanilla_spider_ball": 26,
        }
        for name, index in allow_vanilla.items():
            if getattr(configuration, name):
                enabled_resources.add(index)

        if configuration.elevators.is_vanilla:
            # Vanilla Great Temple Emerald Translator Gate
            enabled_resources.add(18)

        if configuration.safe_zone.prevents_dark_aether:
            # Safe Zone
            enabled_resources.add(2)

    return {
        resource: 1 if resource.index in enabled_resources else 0
        for resource in resource_database.misc
    }


def prime1_progressive_damage_reduction(db: ResourceDatabase, current_resources: CurrentResources):
    num_suits = sum(current_resources.get(db.get_item_by_name(suit), 0)
                    for suit in ["Varia Suit", "Gravity Suit", "Phazon Suit"])
    if num_suits >= 3:
        return 0.5
    elif num_suits == 2:
        return 0.8
    elif num_suits == 1:
        return 0.9
    else:
        return 1


def prime1_absolute_damage_reduction(db: ResourceDatabase, current_resources: CurrentResources):
    if current_resources.get(db.get_item_by_name("Phazon Suit"), 0) > 0:
        return 0.5
    elif current_resources.get(db.get_item_by_name("Gravity Suit"), 0) > 0:
        return 0.8
    elif current_resources.get(db.get_item_by_name("Varia Suit"), 0) > 0:
        return 0.9
    else:
        return 1


def patch_resource_database(db: ResourceDatabase, configuration: AnyGameConfiguration) -> ResourceDatabase:
    base_damage_reduction = db.base_damage_reduction
    damage_reductions = copy.copy(db.damage_reductions)
    requirement_template = copy.copy(db.requirement_template)

    if configuration.game == RandovaniaGame.METROID_PRIME:
        suits = [db.get_item_by_name("Varia Suit")]
        if configuration.heat_protection_only_varia:
            requirement_template["Heat-Resisting Suit"] = ResourceRequirement(db.get_item_by_name("Varia Suit"),
                                                                              1, False)
        else:
            suits.extend([db.get_item_by_name("Gravity Suit"), db.get_item_by_name("Phazon Suit")])

        reductions = [DamageReduction(None, configuration.heat_damage / 10.0)]
        reductions.extend([DamageReduction(suit, 0) for suit in suits])
        damage_reductions[db.get_by_type_and_index(ResourceType.DAMAGE, 5)] = reductions

        if configuration.progressive_damage_reduction:
            base_damage_reduction = prime1_progressive_damage_reduction
        else:
            base_damage_reduction = prime1_absolute_damage_reduction

    elif configuration.game == RandovaniaGame.METROID_PRIME_ECHOES:
        damage_reductions[db.get_by_type_and_index(ResourceType.DAMAGE, 2)] = [
            DamageReduction(None, configuration.varia_suit_damage / 6.0),
            DamageReduction(db.get_item_by_name("Dark Suit"), configuration.dark_suit_damage / 6.0),
            DamageReduction(db.get_item_by_name("Light Suit"), 0.0),
        ]

    return dataclasses.replace(db, damage_reductions=damage_reductions, base_damage_reduction=base_damage_reduction,
                               requirement_template=requirement_template)


def logic_bootstrap(configuration: AnyGameConfiguration,
                    game: GameDescription,
                    patches: GamePatches,
                    ) -> Tuple[GameDescription, State]:
    """
    Core code for starting a new Logic/State.
    :param configuration:
    :param game:
    :param patches:
    :return:
    """
    if not game.mutable:
        raise ValueError("Running logic_bootstrap with non-mutable game")

    # FIXME
    energy_per_tank = getattr(configuration, "energy_per_tank", 100)
    starting_state = calculate_starting_state(game, patches, energy_per_tank)

    if configuration.trick_level.minimal_logic:
        _add_minimal_logic_initial_resources(starting_state.resources,
                                             game,
                                             configuration.major_items_configuration)

    static_resources = trick_resources_for_configuration(configuration.trick_level,
                                                         game.resource_database)
    static_resources.update(version_resources_for_game(game.resource_database))
    static_resources.update(misc_resources_for_configuration(configuration, game.resource_database))

    for resource, quantity in static_resources.items():
        starting_state.resources[resource] = quantity

    game.patch_requirements(starting_state.resources, configuration.damage_strictness.value)

    return game, starting_state



import dataclasses
from enum import Enum

from randovania.bitpacking.bitpacking import BitPackEnum
from randovania.games.game import RandovaniaGame
from randovania.layout.base.base_configuration import BaseConfiguration
from randovania.layout.prime1.artifact_mode import LayoutArtifactMode
from randovania.layout.lib.teleporters import TeleporterConfiguration


class LayoutCutsceneMode(BitPackEnum, Enum):
    ORIGINAL = "original"
    COMPETITIVE = "competitive"
    MINOR = "minor"
    MAJOR = "major"


@dataclasses.dataclass(frozen=True)
class PrimeConfiguration(BaseConfiguration):
    elevators: TeleporterConfiguration
    energy_per_tank: int = dataclasses.field(metadata={"min": 1, "max": 1000, "precision": 1})
    artifact_target: LayoutArtifactMode
    artifact_minimum_progression: int = dataclasses.field(metadata={"min": 0, "max": 99})
    heat_damage: float = dataclasses.field(metadata={"min": 0.1, "max": 99.9, "precision": 3.0})
    warp_to_start: bool
    heat_protection_only_varia: bool
    progressive_damage_reduction: bool
    allow_underwater_movement_without_gravity: bool
    small_samus: bool

    main_plaza_door: bool
    backwards_frigate: bool
    backwards_labs: bool
    backwards_upper_mines: bool
    backwards_lower_mines: bool
    phazon_elite_without_dynamo: bool

    qol_game_breaking: bool
    qol_pickup_scans: bool
    qol_cutscenes: LayoutCutsceneMode

    @classmethod
    def game_enum(cls) -> RandovaniaGame:
        return RandovaniaGame.METROID_PRIME



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_starting_area.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *


class Ui_PresetStartingArea(object):
    def setupUi(self, PresetStartingArea):
        if not PresetStartingArea.objectName():
            PresetStartingArea.setObjectName(u"PresetStartingArea")
        PresetStartingArea.resize(505, 463)
        self.centralWidget = QWidget(PresetStartingArea)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.starting_area_layout = QVBoxLayout(self.centralWidget)
        self.starting_area_layout.setSpacing(6)
        self.starting_area_layout.setContentsMargins(11, 11, 11, 11)
        self.starting_area_layout.setObjectName(u"starting_area_layout")
        self.starting_area_layout.setContentsMargins(2, 2, 2, 2)
        self.startingarea_description = QLabel(self.centralWidget)
        self.startingarea_description.setObjectName(u"startingarea_description")
        self.startingarea_description.setWordWrap(True)

        self.starting_area_layout.addWidget(self.startingarea_description)

        self.starting_area_quick_fill_layout = QHBoxLayout()
        self.starting_area_quick_fill_layout.setSpacing(6)
        self.starting_area_quick_fill_layout.setObjectName(u"starting_area_quick_fill_layout")
        self.starting_area_quick_fill_label = QLabel(self.centralWidget)
        self.starting_area_quick_fill_label.setObjectName(u"starting_area_quick_fill_label")

        self.starting_area_quick_fill_layout.addWidget(self.starting_area_quick_fill_label)

        self.starting_area_quick_fill_ship = QPushButton(self.centralWidget)
        self.starting_area_quick_fill_ship.setObjectName(u"starting_area_quick_fill_ship")

        self.starting_area_quick_fill_layout.addWidget(self.starting_area_quick_fill_ship)

        self.starting_area_quick_fill_save_station = QPushButton(self.centralWidget)
        self.starting_area_quick_fill_save_station.setObjectName(u"starting_area_quick_fill_save_station")

        self.starting_area_quick_fill_layout.addWidget(self.starting_area_quick_fill_save_station)


        self.starting_area_layout.addLayout(self.starting_area_quick_fill_layout)

        self.starting_locations_area = QScrollArea(self.centralWidget)
        self.starting_locations_area.setObjectName(u"starting_locations_area")
        self.starting_locations_area.setWidgetResizable(True)
        self.starting_locations_contents = QWidget()
        self.starting_locations_contents.setObjectName(u"starting_locations_contents")
        self.starting_locations_contents.setGeometry(QRect(0, 0, 499, 313))
        self.starting_locations_layout = QGridLayout(self.starting_locations_contents)
        self.starting_locations_layout.setSpacing(6)
        self.starting_locations_layout.setContentsMargins(11, 11, 11, 11)
        self.starting_locations_layout.setObjectName(u"starting_locations_layout")
        self.starting_locations_layout.setContentsMargins(4, 4, 4, 4)
        self.starting_locations_area.setWidget(self.starting_locations_contents)

        self.starting_area_layout.addWidget(self.starting_locations_area)

        PresetStartingArea.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetStartingArea)

        QMetaObject.connectSlotsByName(PresetStartingArea)
    # setupUi

    def retranslateUi(self, PresetStartingArea):
        PresetStartingArea.setWindowTitle(QCoreApplication.translate("PresetStartingArea", u"Starting Area", None))
        self.startingarea_description.setText(QCoreApplication.translate("PresetStartingArea", u"<html><head/><body><p>The area where the game starts at can be customized, being selected randomly from a list.</p><p>For ease of use, you can select some pre-defined list of locations. They are:<br/>Ship: Just Temple Grounds - Landing Site, the vanilla location.<br/>Save Stations: All areas with Save Stations.</p><p><span style=\" font-weight:600;\">WARNING</span>: depending on the starting items that are configured, it may be impossible to start at the chosen place. In that case, the generation will fail.</p></body></html>", None))
        self.starting_area_quick_fill_label.setText(QCoreApplication.translate("PresetStartingArea", u"Quick Fill with:", None))
        self.starting_area_quick_fill_ship.setText(QCoreApplication.translate("PresetStartingArea", u"Ship (Default)", None))
        self.starting_area_quick_fill_save_station.setText(QCoreApplication.translate("PresetStartingArea", u"Save Station", None))
    # retranslateUi




CUSTOM_NAMES = {
    # Great Temple
    408633584: "Temple Transport Emerald",
    2399252740: "Temple Transport Violet",
    2556480432: "Temple Transport Amber",

    # Temple Grounds to Great Temple
    1345979968: "Sanctuary Quadrant",
    1287880522: "Agon Quadrant",
    2918020398: "Torvus Quadrant",

    # Temple Grounds to Areas
    1660916974: "Agon Gate",
    2889020216: "Torvus Gate",
    3455543403: "Sanctuary Gate",

    # Agon
    1473133138: "Agon Entrance",
    2806956034: "Agon Portal Access",
    3331021649: "Agon Temple Access",

    # Torvus
    1868895730: "Torvus Entrance",
    3479543630: "Torvus Temple Access",
    3205424168: "Lower Torvus Access",

    # Sanctuary
    3528156989: "Sanctuary Entrance",
    900285955: "Sanctuary Spider side",
    3145160350: "Sanctuary Vault side"
}


from randovania.gui.generated.preset_echoes_patches_ui import Ui_PresetEchoesPatches
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.preset import Preset


class PresetEchoesPatches(PresetTab, Ui_PresetEchoesPatches):

    def __init__(self, editor: PresetEditor):
        super().__init__(editor)
        self.setupUi(self)

        self.include_menu_mod_label.setText(self.include_menu_mod_label.text().replace("color:#0000ff;", ""))

        # Signals
        self.warp_to_start_check.stateChanged.connect(self._persist_option_then_notify("warp_to_start"))
        self.include_menu_mod_check.stateChanged.connect(self._persist_option_then_notify("include_menu_mod"))

    @property
    def uses_patches_tab(self) -> bool:
        return True

    def on_preset_changed(self, preset: Preset):
        config = preset.configuration
        self.warp_to_start_check.setChecked(config.warp_to_start)
        self.include_menu_mod_check.setChecked(config.menu_mod)



from PySide2 import QtCore

from randovania.gui.generated.preset_prime_goal_ui import Ui_PresetPrimeGoal
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.preset import Preset
from randovania.layout.prime1.artifact_mode import LayoutArtifactMode


class PresetPrimeGoal(PresetTab, Ui_PresetPrimeGoal):

    def __init__(self, editor: PresetEditor):
        super().__init__(editor)
        self.setupUi(self)

        self.goal_layout.setAlignment(QtCore.Qt.AlignTop)
        self.slider.valueChanged.connect(self._on_slider_changed)
        self.min_progression_spin.valueChanged.connect(self._on_spin_changed)

    @property
    def uses_patches_tab(self) -> bool:
        return False

    def _update_editor(self):
        with self._editor as editor:
            editor.set_configuration_field(
                "artifact_target",
                LayoutArtifactMode(self.slider.value())
            )

    def _on_spin_changed(self):
        with self._editor as editor:
            editor.set_configuration_field(
                "artifact_minimum_progression",
                self.min_progression_spin.value(),
            )

    def _on_slider_changed(self):
        self.slider_label.setText(str(self.slider.value()))
        self._update_editor()

    def on_preset_changed(self, preset: Preset):
        artifacts = preset.configuration.artifact_target
        self.slider.setValue(artifacts.num_artifacts)
        self.min_progression_spin.setValue(preset.configuration.artifact_minimum_progression)



import struct
from typing import Optional, List, Tuple

from randovania.dol_patching import assembler
from randovania.game_connection.connection_base import Inventory
from randovania.game_connection.connector.prime_remote_connector import PrimeRemoteConnector
from randovania.game_connection.executor.memory_operation import MemoryOperation, MemoryOperationExecutor
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.world.world import World
from randovania.games.prime import (all_prime_dol_patches, prime_items, prime1_dol_patches)
from randovania.games.prime.prime1_dol_patches import Prime1DolVersion


def format_received_item(item_name: str, player_name: str) -> str:
    special = {
        "Locked Power Bomb Expansion": ("Received Power Bomb Expansion from {provider_name}, "
                                        "but the main Power Bomb is required to use it."),
        "Locked Missile Expansion": ("Received Missile Expansion from {provider_name}, "
                                     "but the Missile Launcher is required to use it."),
    }

    generic = "Received {item_name} from {provider_name}."

    return special.get(item_name, generic).format(item_name=item_name, provider_name=player_name)


def _prime1_powerup_offset(item_index: int) -> int:
    powerups_offset = 0x24
    vector_data_offset = 0x4
    powerup_size = 0x8
    return (powerups_offset + vector_data_offset) + (item_index * powerup_size)


class Prime1RemoteConnector(PrimeRemoteConnector):
    version: Prime1DolVersion

    def __init__(self, version: Prime1DolVersion):
        super().__init__(version)

    def _asset_id_format(self):
        return ">I"

    async def current_game_status(self, executor: MemoryOperationExecutor) -> Tuple[bool, Optional[World]]:
        """
        Fetches the world the player's currently at, or None if they're not in-game.
        :param executor:
        :return: bool indicating if there's a pending `execute_remote_patches` operation.
        """

        cstate_manager_global = self.version.cstate_manager_global

        asset_id_size = struct.calcsize(self._asset_id_format())
        mlvl_offset = 0x84
        cplayer_offset = 0x84c

        memory_ops = [
            MemoryOperation(self.version.game_state_pointer, offset=mlvl_offset, read_byte_count=asset_id_size),
            MemoryOperation(cstate_manager_global + 0x2, read_byte_count=1),
            MemoryOperation(cstate_manager_global + cplayer_offset, offset=0, read_byte_count=4),
        ]
        results = await executor.perform_memory_operations(memory_ops)

        pending_op_byte = results[memory_ops[1]]

        has_pending_op = pending_op_byte != b"\x00"
        return has_pending_op, self._current_status_world(results.get(memory_ops[0]),
                                                          results.get(memory_ops[2]))

    async def _memory_op_for_items(self, executor: MemoryOperationExecutor, items: List[ItemResourceInfo],
                                   ) -> List[MemoryOperation]:
        player_state_pointer = int.from_bytes(await executor.perform_single_memory_operation(MemoryOperation(
            address=self.version.cstate_manager_global + 0x8b8,
            read_byte_count=4,
        )), "big")
        return [
            MemoryOperation(
                address=player_state_pointer,
                offset=_prime1_powerup_offset(item.index),
                read_byte_count=8,
            )
            for item in items
        ]

    async def _patches_for_pickup(self, provider_name: str, pickup: PickupEntry, inventory: Inventory
                                  ) -> Tuple[List[List[assembler.BaseInstruction]], str]:
        item_name, resources_to_give = self._resources_to_give_for_pickup(pickup, inventory)

        self.logger.debug(f"Resource changes for {pickup.name} from {provider_name}: {resources_to_give}")

        patches: List[List[assembler.BaseInstruction]] = []

        for item, delta in resources_to_give.items():
            if delta == 0:
                continue

            if item.index not in prime_items.ARTIFACT_ITEMS:
                patches.append(all_prime_dol_patches.adjust_item_amount_and_capacity_patch(
                    self.version.powerup_functions, self.game.game, item.index, delta,
                ))
            else:
                if item.index > 29:
                    layer_id = item.index - 28
                else:
                    layer_id = 24  # Truth layer

                patches.append(all_prime_dol_patches.increment_item_capacity_patch(
                    self.version.powerup_functions, self.game.game, item.index, delta
                ))
                patches.append(prime1_dol_patches.set_artifact_layer_active_patch(
                    self.version, layer_id, delta > 0
                ))

        return patches, format_received_item(item_name, provider_name)

    def _write_string_to_game_buffer(self, message: str) -> MemoryOperation:
        return super()._write_string_to_game_buffer("&just=center;" + message)



from randovania.games.game import RandovaniaGame
from randovania.games.prime import prime1_elevators, echoes_elevators
from randovania.game_description.world.area_location import AreaLocation
from randovania.game_description.world.world_list import WorldList

SHORT_CUSTOM_NAMES_FOR_ELEVATORS = {
    RandovaniaGame.METROID_PRIME: prime1_elevators.SHORT_UI_CUSTOM_NAMES,
    RandovaniaGame.METROID_PRIME_ECHOES: echoes_elevators.CUSTOM_NAMES
}

CUSTOM_NAMES_FOR_ELEVATORS = {
    RandovaniaGame.METROID_PRIME: prime1_elevators.UI_CUSTOM_NAMES,
    RandovaniaGame.METROID_PRIME_ECHOES: echoes_elevators.CUSTOM_NAMES
}

def get_elevator_name_or_default(
    game: RandovaniaGame, 
    asset_id: int, 
    default: str
) -> str:
    return CUSTOM_NAMES_FOR_ELEVATORS.get(game, {}).get(asset_id, default)


def get_elevator_or_area_name(
        game: RandovaniaGame,
        world_list: WorldList,
        area_location: AreaLocation,
        include_world_name: bool
) -> str:
    return _get_elevator_or_area_name(CUSTOM_NAMES_FOR_ELEVATORS, game, world_list, area_location, include_world_name)

def get_short_elevator_or_area_name(
        game: RandovaniaGame,
        world_list: WorldList,
        area_location: AreaLocation,
        include_world_name: bool
) -> str:
    return _get_elevator_or_area_name(SHORT_CUSTOM_NAMES_FOR_ELEVATORS, game, world_list, area_location, include_world_name)

def _get_elevator_or_area_name(
        custom_names_to_use: dict,
        game: RandovaniaGame,
        world_list: WorldList,
        area_location: AreaLocation,
        include_world_name: bool
) -> str:
    custom_names_by_game = custom_names_to_use.get(game, {})

    if area_location.area_asset_id in custom_names_by_game:
        return custom_names_by_game[area_location.area_asset_id]

    else:
        area = world_list.area_by_area_location(area_location)

        if include_world_name:
            return world_list.area_name(area)
        else:
            return area.name
